from controls import OutputTextVerbatim, Sidebar
from playwright.sync_api import Page
from utils.deploy_utils import create_deploys_app_url_fixture, skip_if_not_chrome
from utils.express_utils import compare_annotations

from shiny import ui
from shiny.express import ui as xui

app_url = create_deploys_app_url_fixture("express_page_sidebar")


@skip_if_not_chrome
def test_express_page_sidebar(page: Page, app_url: str) -> None:
    page.goto(app_url)

    sidebar = Sidebar(page, "sidebar")
    sidebar.expect_text("SidebarTitle Sidebar Content")
    output_txt = OutputTextVerbatim(page, "txt")
    output_txt.expect_value("50")
    compare_annotations(ui.sidebar, xui.sidebar)



import pytest
from conftest import ShinyAppProc
from controls import ValueBox
from playwright.sync_api import Page


@pytest.mark.parametrize("value_box_id", ["valuebox1", "valuebox2"])
def test_valuebox(page: Page, local_app: ShinyAppProc, value_box_id: str) -> None:
    page.goto(local_app.url)

    value_box1 = ValueBox(page, "valuebox1")
    # value_box1.expect_height(None); TODO-fix-karan;
    value_box1.expect_title("KPI Title")
    value_box1.expect_value("$1 Billion Dollars")
    value_box1.expect_full_screen(False)
    value_box1.open_full_screen()
    value_box1.expect_full_screen(True)
    value_box1.expect_body(["30% VS PREVIOUS 30 DAYS"])
    value_box1.close_full_screen()
    value_box1.expect_full_screen(False)

    value_box2 = ValueBox(page, "valuebox2")
    # value_box2.expect_height("300px"); TODO-fix-karan;
    value_box2.expect_title("title")
    value_box2.expect_value("value")
    value_box2.expect_full_screen(False)
    value_box2.open_full_screen()
    value_box2.expect_full_screen(True)
    value_box2.expect_body(["content", "more body"])
    value_box2.close_full_screen()
    value_box2.expect_full_screen(False)

    title_tag_name = (
        value_box2.loc_title.locator("*")
        .nth(0)
        .evaluate("el => el.tagName.toLowerCase()")
    )
    assert title_tag_name == "p"



# Experimental

from ._card import (
    WrapperCallable,
    ImgContainer,
    card,
    card_body,
    card_image,
    card_title,
)

from ._deprecated import (
    # Input Switch
    toggle_switch,
    # Input Text Area
    input_text_area,
    # Navs
    navset_pill_card,
    navset_tab_card,
    # Sidebar
    Sidebar,
    sidebar,
    layout_sidebar,
    toggle_sidebar,
    sidebar_toggle,
    panel_sidebar,
    panel_main,
    DeprecatedPanelSidebar,
    DeprecatedPanelMain,
    # Tooltip
    tooltip_update,
    tooltip_toggle,
    tooltip,
    toggle_tooltip,
    update_tooltip,
    # Css
    CssUnit,
    as_css_padding,
    as_css_unit,
    as_width_unit,
    # Popover
    popover,
    toggle_popover,
    update_popover,
    # Accordion
    AccordionPanel,
    accordion,
    accordion_panel,
    accordion_panel_close,
    accordion_panel_insert,
    accordion_panel_open,
    accordion_panel_remove,
    accordion_panel_set,
    update_accordion_panel,
    # Fill
    as_fill_carrier,
    as_fill_item,
    as_fillable_container,
    is_fill_carrier,
    is_fill_item,
    is_fillable_container,
    remove_all_fill,
    # Card
    TagCallable,
    CardItem,
    card_footer,
    card_header,
    # Value Box
    showcase_left_center,
    showcase_top_right,
    value_box,
    # Layout
    layout_column_wrap,
    # Navs
    navset_bar,
    navset_card_pill,
    navset_card_tab,
    # Outputs
    output_image,
    output_plot,
    output_ui,
    # Page
    page_fillable,
    page_navbar,
    page_sidebar,
)


__all__ = (
    # Card
    "WrapperCallable",
    "ImgContainer",
    "TagCallable",
    "card",
    "card_title",
    "card_body",
    "card_image",
    # Deprecated
    # # Sidebar
    "Sidebar",
    "sidebar",
    "layout_sidebar",
    "toggle_sidebar",
    "sidebar_toggle",
    "panel_sidebar",
    "panel_main",
    "DeprecatedPanelSidebar",
    "DeprecatedPanelMain",
    # # Input Switch
    "toggle_switch",
    # # Input Text Area
    "input_text_area",
    # # Navs
    "navset_pill_card",
    "navset_tab_card",
    # # Tooltip
    "tooltip",
    "toggle_tooltip",
    "update_tooltip",
    "tooltip_update",
    "tooltip_toggle",
    # # Sidebar
    "sidebar_toggle",
    "panel_sidebar",
    "panel_main",
    "DeprecatedPanelSidebar",
    "DeprecatedPanelMain",
    # # Css
    "CssUnit",
    "as_css_unit",
    "as_css_padding",
    "as_width_unit",
    # # Popover
    "popover",
    "toggle_popover",
    "update_popover",
    # Accordion
    "AccordionPanel",
    "accordion",
    "accordion_panel",
    "accordion_panel_set",
    "accordion_panel_open",
    "accordion_panel_close",
    "accordion_panel_insert",
    "accordion_panel_remove",
    "update_accordion_panel",
    # # Fill
    "as_fill_carrier",
    "as_fillable_container",
    "as_fill_item",
    "remove_all_fill",
    "is_fill_carrier",
    "is_fillable_container",
    "is_fill_item",
    # # Card
    "TagCallable",
    "CardItem",
    "card_header",
    "card_footer",
    # # ValueBox
    "showcase_left_center",
    "showcase_top_right",
    "value_box",
    # # Layout
    "layout_column_wrap",
    # # Navs
    "navset_bar",
    "navset_card_tab",
    "navset_card_pill",
    # # Output
    "output_image",
    "output_plot",
    "output_ui",
    # # Page
    "page_sidebar",
    "page_fillable",
    "page_navbar",
)



from datetime import date

from shiny.express import ui

ui.input_date("date1", "Date:", value="2016-02-29")
# Default value is the date in client's time zone
ui.input_date("date2", "Date:")
# value is always yyyy-mm-dd, even if the display format is different
ui.input_date("date3", "Date:", value="2016-02-29", format="mm/dd/yy")
# Pass in a Date object
ui.input_date("date4", "Date:", value=date(2016, 2, 29))
# Use different language and different first day of week
ui.input_date("date5", "Date:", language="ru", weekstart=1)
# Start with decade view instead of default month view
ui.input_date("date6", "Date:", startview="decade")
# Disable Mondays and Tuesdays.
ui.input_date("date7", "Date:", daysofweekdisabled=[1, 2])
# Disable specific dates.
ui.input_date(
    "date8",
    "Date:",
    value="2016-02-29",
    datesdisabled=["2016-03-01", "2016-03-02"],
)



from __future__ import annotations

import textwrap
from typing import Awaitable, Callable, cast

from .._namespaces import Root
from ..session import Inputs, Outputs, Session

all = ("MockSession",)


# A very bare-bones mock session class that is used only in shiny.express's UI rendering
# phase.
class MockSession:
    def __init__(self):
        self.ns = Root
        self.input = Inputs({})
        self.output = Outputs(cast(Session, self), self.ns, {}, {})

    # This is needed so that Outputs don't throw an error.
    def _is_hidden(self, name: str) -> bool:
        return False

    def on_ended(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
    ) -> Callable[[], None]:
        return lambda: None

    def __getattr__(self, name: str):
        raise AttributeError(
            textwrap.dedent(
                f"""
            The session attribute `{name}` is not yet available for use. Since this code
            will run again when the session is initialized, you can use `if session:` to
            only run this code when the session is established.
        """
            )
        )



from shiny.express import ui

ui.page_opts(title="Nav Panel Example")

with ui.nav_panel("Page 1"):
    "Page 1 content"

with ui.nav_panel("Page 2"):
    with ui.navset_card_underline():
        with ui.nav_panel("Tab 1"):
            "Tab 1 content"
        with ui.nav_panel("Tab 2"):
            "Tab 2 content"
        with ui.nav_panel("Tab 3"):
            "Tab 3 content"

ui.nav_spacer()

with ui.nav_menu("Links", align="right"):
    with ui.nav_control():
        ui.a("Shiny", href="https://shiny.posit.co/py/", target="_blank")
    "----"
    "Plain text"
    "----"
    with ui.nav_control():
        ui.a("Posit", href="https://posit.co", target="_blank")



from __future__ import annotations

import abc
import json
from typing import TYPE_CHECKING, Any, Literal, Protocol, Union, cast, runtime_checkable

from htmltools import Tag

from .. import ui
from .._docstring import add_example, no_example
from ._dataframe_unsafe import serialize_numpy_dtypes
from .renderer import Jsonifiable, Renderer

if TYPE_CHECKING:
    import pandas as pd


class AbstractTabularData(abc.ABC):
    @abc.abstractmethod
    def to_payload(self) -> Jsonifiable:
        ...


@add_example(ex_dir="../api-examples/data_frame")
class DataGrid(AbstractTabularData):
    """
    Holds the data and options for a :class:`~shiny.render.data_frame` output, for a
    spreadsheet-like view.

    Parameters
    ----------
    data
        A pandas `DataFrame` object, or any object that has a `.to_pandas()` method
        (e.g., a Polars data frame or Arrow table).
    width
        A _maximum_ amount of vertical space for the data grid to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. The
        default is `fit-content`, which sets the grid's width according to its contents.
        Set this to `100%` to use the maximum available horizontal space.
    height
        A _maximum_ amount of vertical space for the data grid to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. If there
        are more rows than can fit in this space, the grid will scroll. Set the height
        to `None` to allow the grid to grow to fit all of the rows (this is not
        recommended for large data sets, as it may crash the browser).
    summary
        If `True` (the default), shows a message like "Viewing rows 1 through 10 of 20"
        below the grid when not all of the rows are being shown. If `False`, the message
        is not displayed. You can also specify a string template to customize the
        message, containing `{start}`, `{end}`, and `{total}` tokens. For example:
        `"Viendo filas {start} a {end} de {total}"`.
    filters
        If `True`, shows a row of filter inputs below the headers, one for each column.
    row_selection_mode
        Use `"none"` to disable row selection, `"single"` to allow a single row to be
        selected at a time, and `"multiple"` to allow multiple rows to be selected by
        clicking on them individually.

    Returns
    -------
    :
        An object suitable for being returned from a `@render.data_frame`-decorated
        output function.

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame`
    * :class:`~shiny.render.data_frame`
    * :class:`~shiny.render.DataTable`
    """

    def __init__(
        self,
        data: object,
        *,
        width: str | float | None = "fit-content",
        height: Union[str, float, None] = "500px",
        summary: Union[bool, str] = True,
        filters: bool = False,
        row_selection_mode: Literal["none", "single", "multiple"] = "none",
    ):
        import pandas as pd

        self.data: pd.DataFrame = cast(
            pd.DataFrame,
            cast_to_pandas(
                data,
                "The DataGrid() constructor didn't expect a 'data' argument of type",
            ),
        )

        self.width = width
        self.height = height
        self.summary = summary
        self.filters = filters
        self.row_selection_mode = row_selection_mode

    def to_payload(self) -> Jsonifiable:
        res = serialize_pandas_df(self.data)
        res["options"] = dict(
            width=self.width,
            height=self.height,
            summary=self.summary,
            filters=self.filters,
            row_selection_mode=self.row_selection_mode,
            style="grid",
        )
        return res


@no_example()
class DataTable(AbstractTabularData):
    """
    Holds the data and options for a :class:`~shiny.render.data_frame` output, for a
    spreadsheet-like view.

    Parameters
    ----------
    data
        A pandas `DataFrame` object, or any object that has a `.to_pandas()` method
        (e.g., a Polars data frame or Arrow table).
    width
        A _maximum_ amount of vertical space for the data table to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. The
        default is `fit-content`, which sets the table's width according to its
        contents. Set this to `100%` to use the maximum available horizontal space.
    height
        A _maximum_ amount of vertical space for the data table to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. If there
        are more rows than can fit in this space, the table body will scroll. Set the
        height to `None` to allow the table to grow to fit all of the rows (this is not
        recommended for large data sets, as it may crash the browser).
    summary
        If `True` (the default), shows a message like "Viewing rows 1 through 10 of 20"
        below the grid when not all of the rows are being shown. If `False`, the message
        is not displayed. You can also specify a string template to customize the
        message, containing `{start}`, `{end}`, and `{total}` tokens. For example:
        `"Viendo filas {start} a {end} de {total}"`.
    filters
        If `True`, shows a row of filter inputs below the headers, one for each column.
    row_selection_mode
        Use `"none"` to disable row selection, `"single"` to allow a single row to be
        selected at a time, and `"multiple"` to allow multiple rows to be selected by
        clicking on them individually.

    Returns
    -------
    :
        An object suitable for being returned from a `@render.data_frame`-decorated
        output function.

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame`
    * :class:`~shiny.render.data_frame`
    * :class:`~shiny.render.DataGrid`
    """

    def __init__(
        self,
        data: object,
        *,
        width: Union[str, float, None] = "fit-content",
        height: Union[str, float, None] = "500px",
        summary: Union[bool, str] = True,
        filters: bool = False,
        row_selection_mode: Union[
            Literal["none"], Literal["single"], Literal["multiple"]
        ] = "none",
    ):
        import pandas as pd

        self.data: pd.DataFrame = cast(
            pd.DataFrame,
            cast_to_pandas(
                data,
                "The DataTable() constructor didn't expect a 'data' argument of type",
            ),
        )

        self.width = width
        self.height = height
        self.summary = summary
        self.filters = filters
        self.row_selection_mode = row_selection_mode

    def to_payload(self) -> Jsonifiable:
        res = serialize_pandas_df(self.data)
        res["options"] = dict(
            width=self.width,
            height=self.height,
            summary=self.summary,
            filters=self.filters,
            row_selection_mode=self.row_selection_mode,
            style="table",
        )
        return res


def serialize_pandas_df(df: "pd.DataFrame") -> dict[str, Any]:
    # Currently, we don't make use of the index; drop it so we don't error trying to
    # serialize it or something
    df = df.reset_index(drop=True)

    res = json.loads(
        # {index: [index], columns: [columns], data: [values]}
        df.to_json(None, orient="split")  # pyright: ignore[reportUnknownMemberType]
    )

    res["type_hints"] = serialize_numpy_dtypes(df)

    return res


DataFrameResult = Union[None, "pd.DataFrame", DataGrid, DataTable]


@add_example()
class data_frame(Renderer[DataFrameResult]):
    """
    Decorator for a function that returns a pandas `DataFrame` object (or similar) to
    render as an interactive table or grid. Features fast virtualized scrolling, sorting,
    filtering, and row selection (single or multiple).

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A :class:`~shiny.render.DataGrid` or :class:`~shiny.render.DataTable` object,
           which can be used to customize the appearance and behavior of the data frame
           output.
        2. A pandas `DataFrame` object. (Equivalent to
           `shiny.render.DataGrid(df)`.)
        3. Any object that has a `.to_pandas()` method (e.g., a Polars data frame or
           Arrow table). (Equivalent to `shiny.render.DataGrid(df.to_pandas())`.)

    Row selection
    -------------
    When using the row selection feature, you can access the selected rows by using the
    `input.<id>_selected_rows()` function, where `<id>` is the `id` of the
    :func:`~shiny.ui.output_data_frame`. The value returned will be `None` if no rows
    are selected, or a tuple of integers representing the indices of the selected rows.
    To filter a pandas data frame down to the selected rows, use
    `df.iloc[list(input.<id>_selected_rows())]`.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator (if that
    decorator is used). Also, the name of the decorated function (or
    ``@output(id=...)``) should match the ``id`` of a :func:`~shiny.ui.output_table`
    container (see :func:`~shiny.ui.output_table` for example usage).

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame`
    * :class:`~shiny.render.DataGrid` and :class:`~shiny.render.DataTable` are the
      objects you can return from the rendering function to specify options.
    """

    def auto_output_ui(self) -> Tag:
        return ui.output_data_frame(id=self.output_id)

    async def transform(self, value: DataFrameResult) -> Jsonifiable:
        if not isinstance(value, AbstractTabularData):
            value = DataGrid(
                cast_to_pandas(
                    value,
                    "@render.data_frame doesn't know how to render objects of type",
                )
            )
        return value.to_payload()


@runtime_checkable
class PandasCompatible(Protocol):
    # Signature doesn't matter, runtime_checkable won't look at it anyway
    def to_pandas(self) -> object:
        ...


def cast_to_pandas(x: object, error_message_begin: str) -> object:
    import pandas as pd

    if not isinstance(x, pd.DataFrame):
        if not isinstance(x, PandasCompatible):
            raise TypeError(
                error_message_begin
                + f" '{str(type(x))}'. Use either a pandas.DataFrame, or an object"
                " that has a .to_pandas() method."
            )
        return x.to_pandas()
    return x



from __future__ import annotations

import datetime
import functools
import logging
import subprocess
import sys
import threading
import time
from contextlib import contextmanager
from pathlib import PurePath
from time import sleep
from types import TracebackType
from typing import (
    IO,
    Any,
    Callable,
    Generator,
    List,
    Literal,
    Optional,
    TextIO,
    Type,
    Union,
)

import pytest

import shiny._utils

__all__ = (
    "ShinyAppProc",
    "create_app_fixture",
    "create_doc_example_core_fixture",
    "create_example_fixture",
    "local_app",
    "run_shiny_app",
    "expect_to_change",
    "retry_with_timeout",
)

here = PurePath(__file__).parent
here_root = here.parent.parent


class OutputStream:
    """Designed to wrap an IO[str] and accumulate the output using a bg thread

    Also allows for blocking waits for particular lines."""

    def __init__(self, io: IO[str], desc: Optional[str] = None):
        self._io = io
        self._closed = False
        self._lines: List[str] = []
        self._cond = threading.Condition()
        self._thread = threading.Thread(
            group=None, target=self._run, daemon=True, name=desc
        )

        self._thread.start()

    def _run(self):
        """Pump lines into self._lines in a tight loop."""

        try:
            while not self._io.closed:
                try:
                    line = self._io.readline()
                except ValueError:
                    # This is raised when the stream is closed
                    break
                if line != "":
                    with self._cond:
                        self._lines.append(line)
                        self._cond.notify_all()
        finally:
            # If we got here, we're finished reading self._io and need to signal any
            # waiters that we're done and they'll never hear from us again.
            with self._cond:
                self._closed = True
                self._cond.notify_all()

    def wait_for(self, predicate: Callable[[str], bool], timeoutSecs: float) -> bool:
        timeoutAt = datetime.datetime.now() + datetime.timedelta(seconds=timeoutSecs)
        pos = 0
        with self._cond:
            while True:
                while pos < len(self._lines):
                    if predicate(self._lines[pos]):
                        return True
                    pos += 1
                if self._closed:
                    return False
                else:
                    remaining = (timeoutAt - datetime.datetime.now()).total_seconds()
                    if remaining < 0 or not self._cond.wait(timeout=remaining):
                        # Timed out
                        raise TimeoutError(
                            "Timeout while waiting for Shiny app to become ready"
                        )

    def __str__(self):
        with self._cond:
            return "".join(self._lines)


def dummyio() -> TextIO:
    io = TextIO()
    io.close()
    return io


class ShinyAppProc:
    def __init__(self, proc: subprocess.Popen[str], port: int):
        self.proc = proc
        self.port = port
        self.url = f"http://127.0.0.1:{port}/"
        self.stdout = OutputStream(proc.stdout or dummyio())
        self.stderr = OutputStream(proc.stderr or dummyio())
        threading.Thread(group=None, target=self._run, daemon=True).start()

    def _run(self) -> None:
        self.proc.wait()
        if self.proc.stdout is not None:
            self.proc.stdout.close()
        if self.proc.stderr is not None:
            self.proc.stderr.close()

    def close(self) -> None:
        sleep(0.5)
        self.proc.terminate()

    def __enter__(self) -> ShinyAppProc:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        self.close()

    def wait_until_ready(self, timeoutSecs: float) -> None:
        error_lines: List[str] = []

        def stderr_uvicorn(line: str) -> bool:
            error_lines.append(line)
            return "Uvicorn running on" in line

        if self.stderr.wait_for(stderr_uvicorn, timeoutSecs=timeoutSecs):
            return
        else:
            raise RuntimeError(
                "Shiny app exited without ever becoming ready. Waiting for 'Uvicorn running on' in stderr. Last 20 lines of stderr:\n"
                + "\n".join(error_lines[-20:])
            )


def run_shiny_app(
    app_file: Union[str, PurePath],
    *,
    port: int = 0,
    cwd: Optional[str] = None,
    wait_for_start: bool = True,
    timeout_secs: float = 10,
    bufsize: int = 64 * 1024,
) -> ShinyAppProc:
    if port == 0:
        port = shiny._utils.random_port()

    child = subprocess.Popen(
        [sys.executable, "-m", "shiny", "run", "--port", str(port), str(app_file)],
        bufsize=bufsize,
        executable=sys.executable,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        encoding="utf-8",
    )

    # TODO: Detect early exit

    sa = ShinyAppProc(child, port)
    if wait_for_start:
        sa.wait_until_ready(timeout_secs)
    return sa


def local_app_fixture_gen(app: PurePath | str):
    sa = run_shiny_app(app, wait_for_start=False)
    try:
        with sa:
            sa.wait_until_ready(30)
            yield sa
    finally:
        logging.warning("Application output:\n" + str(sa.stderr))


ScopeName = Literal["session", "package", "module", "class", "function"]


def create_app_fixture(
    app: Union[PurePath, str],
    scope: ScopeName = "module",
):
    @pytest.fixture(scope=scope)
    def fixture_func():
        # Pass through `yield` via `next(...)` call
        # (`yield` must be on same line as `next`!)
        app_gen = local_app_fixture_gen(app)
        yield next(app_gen)

    return fixture_func


def create_example_fixture(
    example_name: str,
    example_file: str = "app.py",
    scope: ScopeName = "module",
):
    """Used to create app fixtures from apps in py-shiny/examples"""
    return create_app_fixture(
        here_root / "examples" / example_name / example_file, scope
    )


def create_doc_example_fixture(
    example_name: str,
    example_file: str = "app.py",
    scope: ScopeName = "module",
):
    """Used to create app fixtures from apps in py-shiny/shiny/api-examples"""
    return create_app_fixture(
        here_root / "shiny/api-examples" / example_name / example_file, scope
    )


def create_doc_example_core_fixture(
    example_name: str,
    scope: ScopeName = "module",
):
    """Used to create app fixtures from ``app-core.py`` example apps in py-shiny/shiny/api-examples"""
    return create_doc_example_fixture(example_name, "app-core.py", scope)


def create_doc_example_express_fixture(
    example_name: str,
    scope: ScopeName = "module",
):
    """Used to create app fixtures from ``app-express.py`` example apps in py-shiny/shiny/api-examples"""
    return create_doc_example_fixture(example_name, "app-express.py", scope)


def x_create_doc_example_fixture(example_name: str, scope: ScopeName = "module"):
    """Used to create app fixtures from apps in py-shiny/shiny/examples"""
    return create_app_fixture(
        here_root / "shiny/experimental/api-examples" / example_name / "app.py", scope
    )


@pytest.fixture(scope="module")
def local_app(request: pytest.FixtureRequest) -> Generator[ShinyAppProc, None, None]:
    app_gen = local_app_fixture_gen(PurePath(request.path).parent / "app.py")
    yield next(app_gen)


@contextmanager
def expect_to_change(
    func: Callable[[], Any], timeoutSecs: float = 10
) -> Generator[None, None, None]:
    """
    Context manager that yields when the value returned by func() changes. Use this
    around code that has a side-effect of changing some state asynchronously (such as
    all browser actions), to prevent moving onto the next step of the test until this
    one has actually taken effect.

    Raises TimeoutError if the value does not change within timeoutSecs.

    Parameters
    ----------
    func
        A function that returns a value. The value returned by this function is
        compared to the value returned by subsequent calls to this function.
    timeoutSecs
        How long to wait for the value to change before raising TimeoutError.

    Example
    -------

        with expect_to_change(lambda: page.locator("#name").value()):
            page.keyboard.send_keys("hello")

    """

    original_value = func()
    yield

    @retry_with_timeout(timeoutSecs)
    def wait_for_change():
        if func() == original_value:
            raise AssertionError("Value did not change")

    wait_for_change()


def retry_with_timeout(timeout: float = 30):
    """
    Decorator that retries a function until 1) it succeeds, 2) fails with a
    non-assertion error, or 3) repeatedly fails with an AssertionError for longer than
    the timeout. If the timeout elapses, the last AssertionError is raised.

    Parameters
    ----------
    timeout
        How long to wait for the function to succeed before raising the last
        AssertionError.

    Returns
    -------
    A decorator that can be applied to a function.

    Example
    -------

        @retry_with_timeout(30)
        def try_to_find_element():
            if not page.locator("#name").exists():
                raise AssertionError("Element not found")

        try_to_find_element()
    """

    def decorator(func: Callable[[], None]) -> Callable[[], None]:
        @functools.wraps(func)
        def wrapper() -> None:
            start = time.time()
            while True:
                try:
                    return func()
                except AssertionError as e:
                    if time.time() - start > timeout:
                        raise e
                    time.sleep(0.1)

        return wrapper

    return decorator



from conftest import ShinyAppProc, create_doc_example_core_fixture
from controls import InputCheckbox, OutputUi
from playwright.sync_api import Page, expect

app = create_doc_example_core_fixture("input_checkbox")


def test_input_checkbox_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    somevalue = InputCheckbox(page, "somevalue")
    # use output_ui
    output_txt = OutputUi(page, "value")
    expect(somevalue.loc_label).to_have_text("Some value")
    somevalue.expect_label("Some value")

    somevalue.expect_checked(False)
    somevalue.expect_width(None)

    output_txt.expect_text("False")

    somevalue.set(True)

    somevalue.expect_checked(True)

    somevalue.toggle()
    somevalue.expect_checked(False)

    somevalue.toggle()
    somevalue.expect_checked(True)

    output_txt.expect_text("True")



from conftest import ShinyAppProc, create_doc_example_core_fixture
from playwright.sync_api import Page, expect

app = create_doc_example_core_fixture("render_express")

EXPECT_TIMEOUT = 30 * 1000


def test_render_express(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)
    expect(page.get_by_text("Text outside of render express call")).to_have_count(
        1, timeout=EXPECT_TIMEOUT
    )
    expect(page.get_by_text("Text inside of render express call")).to_have_count(
        1, timeout=EXPECT_TIMEOUT
    )
    expect(page.get_by_text("Dynamic slider value: 50")).to_have_count(
        1, timeout=EXPECT_TIMEOUT
    )



from controls import OutputDataFrame
from playwright.sync_api import Page
from utils.deploy_utils import create_deploys_app_url_fixture, skip_if_not_chrome

app_url = create_deploys_app_url_fixture("shiny-express-dataframe")


@skip_if_not_chrome
def test_express_dataframe_deploys(page: Page, app_url: str) -> None:
    page.goto(app_url)

    dataframe = OutputDataFrame(page, "sample_data_frame")
    dataframe.expect_n_row(6)



import os
import subprocess
import tempfile

import pytest
from example_apps import get_apps, reruns, reruns_delay, validate_example
from playwright.sync_api import Page


def subprocess_create(
    app_template: str,
    dest_dir: str = "",
    *,
    mode: str = "core",
    package_name: str = "",
):
    cmd = [
        "shiny",
        "create",
        "--template",
        app_template,
        "--mode",
        mode,
        "--dir",
        dest_dir,
        "--package-name",
        package_name,
    ]

    subprocess.run(cmd)
    assert os.path.isdir(dest_dir)

    # Package templates don't have a top-level app.py file
    if package_name == "":
        assert os.path.isfile(f"{dest_dir}/app.py")
    else:
        assert os.path.isfile(f"{dest_dir}/pyproject.toml")

    print("\nCheck for duplicate files")
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(cmd, check=True)


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("ex_app_path", get_apps("shiny/templates/app-templates"))
def test_template_examples(page: Page, ex_app_path: str) -> None:
    validate_example(page, ex_app_path)


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", ["basic-app", "dashboard", "multi-page"])
def test_create_core(app_template: str, page: Page):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir)
        validate_example(page, f"{tmpdir}/app.py")


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", ["basic-app"])
def test_create_express(app_template: str, page: Page):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir, mode="express")
        validate_example(page, f"{tmpdir}/app.py")


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("app_template", ["js-input", "js-output", "js-react"])
def test_create_js(app_template: str):
    with tempfile.TemporaryDirectory("example_apps") as tmpdir:
        subprocess_create(app_template, dest_dir=tmpdir, package_name="my_component")
        # TODO-Karan: Add test to validate once flag to install packages is implemented



from __future__ import annotations

import json
import os
import subprocess
from typing import Any, Callable, TypeVar

import pytest
import requests
from conftest import ScopeName, local_app_fixture_gen

LOCAL_LOCATION = "local"

__all__ = (
    "create_deploys_app_url_fixture",
    "skip_if_not_chrome",
)

# connect
server_url = os.environ.get("DEPLOY_CONNECT_SERVER_URL")
api_key = os.environ.get("DEPLOY_CONNECT_SERVER_API_KEY")
# shinyapps.io
name = os.environ.get("DEPLOY_SHINYAPPS_NAME")
token = os.environ.get("DEPLOY_SHINYAPPS_TOKEN")
secret = os.environ.get("DEPLOY_SHINYAPPS_SECRET")


deploy_locations = ["connect", "shinyapps"]

CallableT = TypeVar("CallableT", bound=Callable[..., Any])


def skip_if_not_chrome(fn: CallableT) -> CallableT:
    # # Keeping commented to allow for easier local debugging
    # import platform
    # fn = pytest.mark.skipif(
    #     platform.python_version_tuple()[:2] != ("3", "10"),
    #     reason="Test requires Python 3.10",
    # )(fn)
    fn = pytest.mark.only_browser("chromium")(fn)

    return fn


def exception_swallower(
    function: Callable[[str, str], str]
) -> Callable[[str, str], str]:
    def wrapper(app_name: str, app_dir: str) -> str:
        runtime_e: Exception | None = None
        try:
            return function(app_name, app_dir)
        except Exception as e:
            runtime_e = e
        if isinstance(runtime_e, Exception):
            raise RuntimeError("Failed to deploy to server.")

    return wrapper


def run_command(cmd: str) -> str:
    output = subprocess.run(
        cmd,
        check=True,
        capture_output=True,
        text=True,
        shell=True,
    )
    return output.stdout


def deploy_to_connect(app_name: str, app_dir: str) -> str:
    if not api_key:
        raise RuntimeError("No api key found. Cannot deploy.")

    # check if connect app is already deployed to avoid duplicates
    connect_server_lookup_command = f"rsconnect content search --server {server_url} --api-key {api_key} --title-contains {app_name}"
    app_details = run_command(connect_server_lookup_command)
    connect_server_deploy = f"rsconnect deploy shiny {app_dir} --server {server_url} --api-key {api_key} --title {app_name} --verbose"
    # only if the app exists do we replace existing app with new version
    if json.loads(app_details):
        app_id = json.loads(app_details)[0]["guid"]
        connect_server_deploy += f" --app-id {app_id}"

    # Deploy to connect server
    run_command(connect_server_deploy)

    # look up content url in connect server once app is deployed
    output = run_command(connect_server_lookup_command)
    url = json.loads(output)[0]["content_url"]
    app_id = json.loads(output)[0]["guid"]
    # change visibility of app to public
    connect_app_url = f"{server_url}/__api__/v1/content/{app_id}"
    payload = '{"access_type":"all"}'
    headers = {
        "Authorization": f"Key {api_key}",
        "Accept": "application/json",
    }
    response = requests.request("PATCH", connect_app_url, headers=headers, data=payload)
    if response.status_code != 200:
        raise RuntimeError("Failed to change visibility of app.")
    return url


quiet_deploy_to_connect = exception_swallower(deploy_to_connect)


# TODO-future: Supress web browser from opening after deploying - https://github.com/rstudio/rsconnect-python/issues/462
def deploy_to_shinyapps(app_name: str, app_dir: str) -> str:
    # Deploy to shinyapps.io
    shinyapps_deploy = f"rsconnect deploy shiny {app_dir} --account {name} --token {token} --secret {secret} --title {app_name} --verbose"
    run_command(shinyapps_deploy)
    return f"https://{name}.shinyapps.io/{app_name}/"


quiet_deploy_to_shinyapps = exception_swallower(deploy_to_shinyapps)


# Since connect parses python packages, we need to get latest version of shiny on HEAD
def write_requirements_txt(app_dir: str) -> None:
    app_requirements_file_path = os.path.join(app_dir, "app_requirements.txt")
    requirements_file_path = os.path.join(app_dir, "requirements.txt")
    git_cmd = subprocess.run(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE)
    git_hash = git_cmd.stdout.decode("utf-8").strip()
    with open(app_requirements_file_path) as f:
        requirements = f.read()
    with open(requirements_file_path, "w") as f:
        f.write(f"{requirements}\n")
        f.write(f"git+https://github.com/posit-dev/py-shiny.git@{git_hash}\n")


def deploy_app(
    app_file_path: str,
    location: str,
    app_name: str,
) -> str:
    should_deploy_apps = os.environ.get("DEPLOY_APPS", "False") == "true"

    if not should_deploy_apps:
        pytest.skip("`DEPLOY_APPS` does not equal `true`")

    run_on_ci = os.environ.get("CI", "False") == "true"
    repo = os.environ.get("GITHUB_REPOSITORY", "unknown")
    branch_name = os.environ.get("GITHUB_HEAD_REF", "unknown")

    if (
        not run_on_ci
        or repo != "posit-dev/py-shiny"
        or not (branch_name.startswith("deploy") or branch_name == "main")
    ):
        pytest.skip("Not on CI or posit-dev/py-shiny repo or deploy* or main branch")
        raise RuntimeError()

    app_dir = os.path.dirname(app_file_path)
    write_requirements_txt(app_dir)

    deployment_function = {
        "connect": quiet_deploy_to_connect,
        "shinyapps": quiet_deploy_to_shinyapps,
    }[location]

    url = deployment_function(app_name, app_dir)
    return url


def create_deploys_app_url_fixture(
    app_name: str,
    scope: ScopeName = "module",
):
    @pytest.fixture(scope=scope, params=[*deploy_locations, LOCAL_LOCATION])
    def fix_fn(request: pytest.FixtureRequest):
        app_file = os.path.join(os.path.dirname(request.path), "app.py")
        deploy_location = request.param

        if deploy_location == LOCAL_LOCATION:
            shinyapp_proc_gen = local_app_fixture_gen(app_file)
            # Return the `url`
            yield next(shinyapp_proc_gen).url
        elif deploy_location in deploy_locations:
            app_url = deploy_app(
                app_file,
                deploy_location,
                app_name,
            )
            yield app_url

        else:
            raise ValueError(
                "Deploy location not a known location: '", deploy_location, "'"
            )

    return fix_fn



from pathlib import Path

import pandas as pd
from modules import data_view_server, data_view_ui, training_server, training_ui

from shiny import App, Inputs, Outputs, Session, reactive, ui

df = pd.read_csv(Path(__file__).parent / "scores.csv")


app_ui = ui.page_navbar(
    training_ui("tab1"),
    data_view_ui("tab2"),
    sidebar=ui.sidebar(
        ui.input_select(
            "account",
            "Account",
            choices=[
                "Berge & Berge",
                "Fritsch & Fritsch",
                "Hintz & Hintz",
                "Mosciski and Sons",
                "Wolff Ltd",
            ],
        ),
        width="300px",
    ),
    header=ui.include_css(Path(__file__).parent / "styles.css"),
    id="tabs",
    title="Monitoring",
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Calc()
    def filtered_data() -> pd.DataFrame:
        return df.loc[df["account"] == input.account()]

    training_server(id="tab1", df=filtered_data)
    data_view_server(id="tab2", df=filtered_data)


app = App(app_ui, server)



from __future__ import annotations

__all__ = (
    "page_sidebar",
    "page_navbar",
    "page_fillable",
    "page_fluid",
    "page_fixed",
    "page_bootstrap",
    "page_auto",
    "page_output",
)

from typing import Any, Callable, Literal, Optional, Sequence, cast

from htmltools import (
    MetadataNode,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    css,
    div,
    head_content,
    tags,
)

from .._docstring import add_example, no_example
from .._namespaces import resolve_id_or_none
from ..types import MISSING, MISSING_TYPE, NavSetArg
from ._bootstrap import panel_title
from ._html_deps_external import bootstrap_deps
from ._html_deps_py_shiny import page_output_dependency
from ._html_deps_shinyverse import components_dependency
from ._navs import NavMenu, NavPanel, navset_bar
from ._sidebar import Sidebar, layout_sidebar
from ._tag import consolidate_attrs
from ._utils import get_window_title
from .css import CssUnit, as_css_padding, as_css_unit
from .fill._fill import as_fillable_container


@add_example()
def page_sidebar(
    sidebar: Sidebar,
    *args: TagChild | TagAttrs,
    title: Optional[str | Tag | TagList] = None,
    fillable: bool = False,
    fillable_mobile: bool = False,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a page with a sidebar and a title.

    Parameters
    ----------
    sidebar
        Content to display in the sidebar.
    *args
        UI elements.
    title
        A title to display at the top of the page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    **kwargs
        Additional attributes passed to :func:`~shiny.ui.layout_sidebar`.

    Returns
    -------
    :
        A UI element.
    """

    if isinstance(title, str):
        title = tags.h1(title, class_="bslib-page-title")

    attrs, children = consolidate_attrs(*args, **kwargs)

    return page_fillable(
        {"class": "bslib-page-sidebar"},
        title,
        layout_sidebar(
            sidebar,
            *children,
            # Make the main area background white instead of the default gray.
            {"style": "--bslib-shiny-preset-main-bg: white;"},
            attrs,
            fillable=fillable,
            border=False,
            border_radius=False,
        ),
        get_window_title(title, window_title=window_title),
        padding=0,
        gap=0,
        lang=lang,
        fillable_mobile=fillable_mobile,
    )


@no_example()
def page_navbar(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    title: Optional[str | Tag | TagList] = None,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    # Only page_navbar gets enhanced treatement for `fillable`
    # If an `*args`'s `data-value` attr string is in `fillable`, then the component is fillable
    fillable: bool | list[str] = False,
    fillable_mobile: bool = False,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal["static-top", "fixed-top", "fixed-bottom"] = "static-top",
    header: Optional[TagChild] = None,
    footer: Optional[TagChild] = None,
    bg: Optional[str] = None,
    inverse: bool = False,
    underline: bool = True,
    collapsible: bool = True,
    fluid: bool = True,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
) -> Tag:
    """
    Create a page with a navbar and a title.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match its
        ``value``).
    sidebar
        A :func:`~shiny.ui.sidebar` component to display on every page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the elements into an expandable menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.nav`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.page_fluid`

    Example
    -------
    See :func:`~shiny.ui.nav`.
    """
    if sidebar is not None and not isinstance(sidebar, Sidebar):
        raise TypeError(
            "`sidebar=` is not a `Sidebar` instance. Use `ui.sidebar(...)` to create one."
        )

    pageClass = "bslib-page-navbar"
    if sidebar is not None:
        pageClass += " has-page-sidebar"

    tagAttrs: TagAttrs = {"class": pageClass}

    page_args = (
        tagAttrs,
        navset_bar(
            *args,
            title=title,
            id=resolve_id_or_none(id),
            selected=selected,
            sidebar=sidebar,
            fillable=fillable,
            gap=gap,
            padding=padding,
            position=position,
            header=header,
            footer=footer,
            bg=bg,
            inverse=inverse,
            underline=underline,
            collapsible=collapsible,
            fluid=fluid,
        ),
        get_window_title(title, window_title=window_title),
    )
    page_kwargs = {
        "title": None,
        "lang": lang,
    }

    # If a sidebar is provided, we want the layout_sidebar(fill = TRUE) component
    # (which is a sibling of the <nav>) to always fill the page
    if fillable is False and sidebar is None:
        return page_bootstrap(
            *page_args,
            **page_kwargs,
        )

    else:
        return page_fillable(
            *page_args,
            fillable_mobile=fillable_mobile,
            padding=0,
            gap=0,
            **page_kwargs,
        )


@no_example()
def page_fillable(
    *args: TagChild | TagAttrs,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fillable_mobile: bool = False,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a fillable page.

    Parameters
    ----------
    *args
        UI elements.
    padding
        Padding to use for the body. See :func:`~shiny.ui.css.as_css_padding`
        for more details.
    fillable_mobile
        Whether or not the page should fill the viewport's height on mobile devices
        (i.e., narrow windows).
    gap
        A CSS length unit passed through :func:`~shiny.ui.css.as_css_unit`
        defining the `gap` (i.e., spacing) between elements provided to `*args`.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_fixed`
    """
    attrs, children = consolidate_attrs(*args, **kwargs)

    style = css(padding=as_css_padding(padding), gap=as_css_unit(gap))

    page = page_bootstrap(
        head_content(tags.style("html { height: 100%; }")),
        {"class": "bslib-page-fill bslib-gap-spacing", "style": style},
        {"class": "bslib-flow-mobile"} if not fillable_mobile else None,
        attrs,
        *children,
        components_dependency(),
        title=title,
        lang=lang,
    )

    # page returns a <html> tag, but we need to make the <body> fillable
    body = page.children[1]
    if not isinstance(body, Tag) or body.name != "body":
        raise ValueError("Expected a <body> tag")

    page.children[1] = as_fillable_container(body)

    return page


@add_example()
def page_fluid(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    **kwargs: str,
) -> Tag:
    """
    Create a fluid page.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    **kwargs
        Attributes on the page level container.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fixed`
    * :func:`~shiny.ui.page_bootstrap`
    * :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container-fluid"}, *args, **kwargs), title=title, lang=lang
    )


@add_example()
def page_fixed(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    **kwargs: str,
) -> Tag:
    """
    Create a fixed page.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    **kwargs
        Attributes on the page level container.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_bootstrap`
    * :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container"}, *args, **kwargs), title=title, lang=lang
    )


# TODO: implement theme (just Bootswatch for now?)
@no_example()
def page_bootstrap(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a Bootstrap UI page container.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    **kwargs
        Attributes on the the `<body>` tag.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_navbar`
    """
    head = tags.title(title) if title else None
    return tags.html(
        tags.head(head),
        tags.body(*bootstrap_deps(), *args, **kwargs),
        lang=lang,
    )


@no_example()
def page_auto(
    *args: TagChild | TagAttrs,
    title: str | MISSING_TYPE = MISSING,
    window_title: str | MISSING_TYPE = MISSING,
    lang: str | MISSING_TYPE = MISSING,
    fillable: bool | MISSING_TYPE = MISSING,
    full_width: bool = False,
    page_fn: Callable[..., Tag] | None = None,
    **kwargs: object,
) -> Tag:
    """
    A page container which automatically decides which page function to use.

    If there is a top-level nav, this will use :func:`~shiny.ui.page_navbar`. If not,
    and there is a top-level sidebar, this will use :func:`~shiny.ui.page_sidebar`.

    If there are neither top-level navs nor sidebars, this will use the ``fillable`` and
    ``full_width`` arguments to determine which page function to use.

    Parameters
    ----------
    *args
        UI elements. These are used to determine which page function to use, and they
        are also passed along to that page function.
    title
        A title shown on the page.
    window_title
        The browser window title. If no value is provided, this will use the value of
        ``title``.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    fillable
        If there is a top-level sidebar or nav, then the value is passed through to the
        :func:`~shiny.ui.page_sidebar` or :func:`~shiny.ui.page_navbar` function.
        Otherwise, if ``True``, use :func:`~shiny.ui.page_fillable`, where the content
        fills the window; if ``False`` (the default), the value of ``full_width`` will
        determine which page function is used.
    full_width
        This has an effect only if there are no sidebars or top-level navs, and
        ``fillable`` is ``False``. If this is ``False`` (the default), use use
        :func:`~shiny.ui.page_fixed`; if ``True``, use :func:`~shiny.ui.page_fillable`.
    page_fn
        The page function to use. If ``None`` (the default), will automatically choose
        one based on the arguments provided. If not ``None``, this will override all
        heuristics for choosing page functions.
    **kwargs
        Additional arguments, which are passed to the page function.

    Returns
    -------
    :
        A UI element.
    """
    if not isinstance(title, MISSING_TYPE):
        kwargs["title"] = title
    if not isinstance(window_title, MISSING_TYPE):
        kwargs["window_title"] = window_title
    if not isinstance(lang, MISSING_TYPE):
        kwargs["lang"] = lang

    # Presence of a top-level nav items and/or sidebar determines the page function
    navs = [x for x in args if isinstance(x, (NavPanel, NavMenu))]
    sidebars = [x for x in args if isinstance(x, Sidebar)]

    nNavs = len(navs)
    nSidebars = len(sidebars)
    if page_fn is None:
        if nNavs == 0:
            if nSidebars == 0:
                if isinstance(fillable, MISSING_TYPE):
                    fillable = False

                if fillable:
                    page_fn = _page_auto_fillable
                elif full_width:
                    page_fn = _page_auto_fluid
                else:
                    page_fn = _page_auto_fixed

            elif nSidebars == 1:
                if not isinstance(fillable, MISSING_TYPE):
                    kwargs["fillable"] = fillable

                # page_sidebar() needs sidebar to be the first arg
                # TODO: Change page_sidebar() to remove `sidebar` and accept a sidebar as a
                # *arg.
                page_fn = page_sidebar
                args = tuple(sidebars + [x for x in args if x not in sidebars])

            else:
                raise NotImplementedError(
                    "Multiple top-level sidebars not allowed. Did you meant to wrap each one in layout_sidebar()?"
                )

        # At least one nav
        else:
            if not isinstance(fillable, MISSING_TYPE):
                kwargs["fillable"] = fillable

            if nSidebars == 0:
                # TODO: what do we do when nArgs != nNavs? Just let page_navbar handle it (i.e. error)?
                page_fn = page_navbar

            elif nSidebars == 1:
                # TODO: change page_navbar() to remove `sidebar` and accept a sidebar as a
                # *arg.
                page_fn = page_navbar
                args = tuple([x for x in args if x not in sidebars])
                kwargs["sidebar"] = sidebars[0]

            else:
                raise NotImplementedError(
                    "Multiple top-level sidebars not allowed in combination with top-level navs."
                )

    # If we got here, _page_fn is not None, but the type checker needs a little help.
    page_fn = cast(Callable[..., Tag], page_fn)
    return page_fn(*args, **kwargs)


# For `page_fillable`, `page_fluid`, and `page_fixed`, the `title` arg sets the window
# title, but doesn't add anything visible on the page.
#
# In contrast, for `page_auto`, the `title` arg adds a title panel to the page, and the
# `window_title` arg sets the window title.
#
# The wrapper functions below provide the `page_auto` interface, where `title` to add a
# title panel to the page, and `window_title` to set the title of the window. If `title`
# is provided but `window_title` is not, then `window_title` is set to the value of
# `title`.
def _page_auto_fillable(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: Any,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fillable(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


def _page_auto_fluid(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: str,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fluid(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


def _page_auto_fixed(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: Any,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fixed(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


@no_example()
def page_output(id: str) -> Tag:
    """
    Create a page container where the entire body is a UI output.

    Parameters
    ----------
    id
        An output id.

    Returns
    -------
    :
        A UI element which is meant to be used as a page container.
    """
    return tags.html(
        tags.head(),
        tags.body(id=id, class_="shiny-page-output"),
        page_output_dependency(),
        lang="en",
    )



import re
import time

from conftest import ShinyAppProc
from controls import InputSlider, InputSliderRange, OutputTextVerbatim
from playwright.sync_api import Page

from shiny.types import MISSING


def test_slider_regular(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s0 = InputSlider(page, "s0")
    s0.expect_label("Integer")
    s0.expect_value("500")
    s0.expect_min("0")
    s0.expect_max("1000")
    s0.expect_step("1")
    s0.expect_ticks("false")
    s0.expect_sep(",")
    s0.expect_pre(None)
    s0.expect_post(None)
    s0.expect_time_format(None)
    s0.expect_timezone(None)
    s0.expect_drag_range(None)
    s0.expect_animate(exists=False)
    OutputTextVerbatim(page, "txt0").expect_value("500")

    new_val = "20"
    s0.set(new_val)
    s0.expect_value(new_val)
    OutputTextVerbatim(page, "txt0").expect_value(new_val)


def test_slider_range(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s1 = InputSliderRange(page, "s1")
    s1.expect_label("Range")
    s1.expect_value(("200", "500"))
    s1.expect_min("1")
    s1.expect_max("1000")
    s1.expect_step("1")
    s1.expect_ticks("false")
    s1.expect_sep(",")
    s1.expect_pre(None)
    s1.expect_post(None)
    s1.expect_time_format(None)
    s1.expect_timezone(None)
    s1.expect_drag_range("true")
    s1.expect_animate(exists=False)
    OutputTextVerbatim(page, "txt1").expect_value("(200, 500)")

    new_val = ("605", "840")
    s1.set(new_val, max_err_values=1000)
    try:
        s1.expect_value((MISSING, MISSING))  # type: ignore
    except ValueError as e:
        assert re.search("tuple entries cannot", str(e))
    s1.expect_value((new_val[0], MISSING))
    s1.expect_value((MISSING, new_val[1]))
    s1.expect_value(new_val)
    OutputTextVerbatim(page, "txt1").expect_value(f"({new_val[0]}, {new_val[1]})")


def test_slider_custom_format(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s2 = InputSlider(page, "s2")
    s2.expect_label("Custom Format")
    s2.expect_value("$0")
    s2.expect_min("0")
    s2.expect_max("10000")
    s2.expect_step("2500")
    s2.expect_ticks("true")
    s2.expect_sep(",")
    s2.expect_pre("$")
    s2.expect_post(None)
    s2.expect_time_format(None)
    s2.expect_timezone(None)
    s2.expect_drag_range(None)
    s2.expect_animate_options(loop=True, interval=500)
    OutputTextVerbatim(page, "txt2").expect_value("0")

    s2.set("$7,500")
    s2.expect_value("$7,500")
    OutputTextVerbatim(page, "txt2").expect_value("7500")


def test_slider_loop(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s3 = InputSlider(page, "s3")
    s3.expect_label("Looping Animation")
    s3.expect_value("1,000")
    s3.expect_min("1")
    s3.expect_max("2000")
    s3.expect_step("10")
    s3.expect_ticks("false")
    s3.expect_sep(",")
    s3.expect_pre(None)
    s3.expect_post(None)
    s3.expect_time_format(None)
    s3.expect_timezone(None)
    s3.expect_drag_range(None)
    s3.expect_animate_options(loop=True, interval=300)
    OutputTextVerbatim(page, "txt3").expect_value("1000")

    s3.set("1,441")
    s3.expect_value("1,441")
    OutputTextVerbatim(page, "txt3").expect_value("1441")

    # Play for a little bit
    s3.click_play()
    time.sleep(400 / 1000)
    s3.click_pause()
    # Make sure the value changed
    try:
        s3.expect_value("1,441", timeout=100)
    except AssertionError:
        pass


def test_slider_play(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s4 = InputSlider(page, "s4")
    s4.expect_label("Single Animation")
    s4.expect_value("1")
    s4.expect_min("0")
    s4.expect_max("5")
    s4.expect_step("1")
    s4.expect_ticks("false")
    s4.expect_sep(",")
    s4.expect_pre(None)
    s4.expect_post(None)
    s4.expect_time_format(None)
    s4.expect_timezone(None)
    s4.expect_drag_range(None)
    s4.expect_animate_options(loop=False, interval=100)
    OutputTextVerbatim(page, "txt4").expect_value("1")

    s4.click_play()
    s4.expect_value("5")
    s4.click_play()  # can click again!
    s4.expect_value("5")



from __future__ import annotations

import base64
import io
import mimetypes
from pathlib import Path, PurePath
from typing import Literal, Optional, Protocol

from htmltools import (
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagFunction,
    Tagifiable,
    css,
    tags,
)

from ..._docstring import add_example
from ...types import MISSING, MISSING_TYPE
from ...ui._card import CardItem, WrapperCallable, _card_impl, card_body
from ...ui.css import CssUnit, as_css_unit
from ...ui.fill import as_fill_item, as_fillable_container

__all__ = (
    # Worried about `wrapper`
    "card",
    # Do not want to expose card_body yet
    "card_body",
    # Questioning:
    "card_title",
    "card_image",
)


def card(
    *args: TagChild | TagAttrs | CardItem,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    wrapper: WrapperCallable | None | MISSING_TYPE = MISSING,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    A Bootstrap card component

    A card is a general purpose container that groups related UI elements together with a border
    and optional padding. To learn more about `card()`s, see [the bslib card
    article](https://rstudio.github.io/bslib/articles/cards.html).

    Parameters
    ----------
    *args
        Unnamed arguments can be any valid child of an :class:`~htmltools.Tag` (This
        includes card items such as :func:`~shiny.experimental.ui.card_body`).
    full_screen
        If `True`, an icon will appear when the user's pointer hovers over the card body. Clicking the
        icon expands the card to fit viewport size.
    height, max_height, min_height
        Any valid CSS unit (e.g., `height="200px"`). These will not apply when a card is made
        `full_screen`. In this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`.
    fill
        Whether or not to allow the card to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    class_
        Additional CSS classes for the returned Tag.
    wrapper
        A function that returns a UI element to call on any unnamed arguments in `*args`
        that are not already card item(s) (like
        :func:`~shiny.ui.card_header`,
        :func:`~shiny.experimental.ui.card_body`, etc.). Note that non-card items are
        grouped together into one `wrapper` call (e.g. given `card("a", "b",
        card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
        `"b"` and once with `"d"`).
    **kwargs
        HTML attributes on the returned Tag.

    Returns
    -------
    :
        A :func:`~shiny.ui.tags.div` tag.

    See Also
    --------
    * :func:`~shiny.ui.layout_column_wrap` for laying out multiple cards
      or multiple columns inside a card.
    * :func:`~shiny.ui.card_header` for creating a header within a card.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within a card body.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside a card.
    * :func:`~shiny.ui.card_footer` for creating a footer within a card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to a card.
    """
    return _card_impl(
        *args,
        full_screen=full_screen,
        height=height,
        max_height=max_height,
        min_height=min_height,
        fill=fill,
        class_=class_,
        wrapper=wrapper,
        **kwargs,
    )


############################################################################


@add_example()
def card_title(
    *args: TagChild | TagAttrs,
    container: TagFunction = tags.h5,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    """
    A card title container

    :func:`~shiny.experimental.ui.card_title` creates a general container for the "title" of
    a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    Parameters
    ----------
    *args
        Contents to appear in the card's title, or tag attributes to pass to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned :class:`~htmltools.Tag` object. Defaults to
        :func:`~shiny.ui.tags.h5`.
    **kwargs
        Additional HTML attributes for the returned :class:`~htmltools.Tag` object.

    Returns
    -------
    :
        An :class:`~htmltools.Tag` object.

    See Also
    --------
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_header` for creating a header within a card.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside a card.
    * :func:`~shiny.ui.card_footer` for creating a footer within a card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to a card.
    """
    return container(*args, **kwargs)


class ImgContainer(Protocol):
    """
    A callable that wraps the return value of `card_image()`. To isolate your object in a card, return a :class:`~shiny.ui.CardItem`.
    """

    def __call__(self, *args: Tag) -> Tagifiable:
        """
        Wraps the return value of `card_image()`.

        Parameters
        ----------
        *args
            The return value of `card_image()`.

        Returns
        -------
        :
            A tagifiable object, such as a :class:`~htmltools.Tag` or
            :class:`~shiny.ui.CardItem` object.
        """
        ...


@add_example()
def card_image(
    file: str | Path | PurePath | io.BytesIO | None,
    *args: TagAttrs,
    href: Optional[str] = None,
    border_radius: Literal["top", "bottom", "all", "none"] = "top",
    mime_type: Optional[str] = None,
    class_: Optional[str] = None,
    height: Optional[CssUnit] = None,
    fill: bool = True,
    width: Optional[CssUnit] = None,
    # Required so that multiple `card_images()` are not put in the same `card()`
    container: ImgContainer = card_body,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    """
    A card image container

    :func:`~shiny.experimental.ui.card_image` creates a general container for an image within a
    :func:`~shiny.ui.card`. This component is designed to be
    provided as a direct child to :func:`~shiny.ui.card`.

    Parameters
    ----------
    file
        A file path pointing to an image. The image will be base64 encoded and provided to
        the `src` attribute of the `<img>` tag. Alternatively, you may set this value to
        `None` and provide the `src` yourself via `*args:TagAttrs` or
        `**kwargs:TagAttrValue` (e.g., `{"src": "HOSTED_PATH_TO_IMAGE"}` or
        `src="HOSTED_PATH_TO_IMAGE"`).
    *args
        A dictionary of tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    href
        An optional URL to link to.
    border_radius
        Where to apply `border-radius` on the image.
    mime_type
        The mime type of the `file`.
    class_
        Additional CSS classes for the resolved :class:`~htmltools.Tag` object.
    height
        Any valid CSS unit (e.g., `height="200px"`). `height` will not apply when a card is made
        `full_screen`. In this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`.
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    width
        Any valid CSS unit (e.g., `width="100%"`).
    container
        Method to wrap the returned :class:`~htmltools.Tag` object. Defaults to
        :func:`~shiny.experimental.ui.card_body`.
        If :func:`~shiny.experimental.ui.card_body` is used, each image will be in separate cards. If
        the `container` method does not return a :class:`~shiny.ui.CardItem`, it
        allows for consecutive non-`CardItem` objects to be bundled into a single
        :func:`~shiny.experimental.ui.card_body` within :func:`~shiny.ui.card`.
    **kwargs
        Additional HTML attributes for the resolved :class:`~htmltools.Tag`.
    """
    src = None
    if file is not None:
        if isinstance(file, io.BytesIO):
            b64_str = base64.b64encode(file.read()).decode("utf-8")
            if mime_type is None:
                raise ValueError(
                    "`mime_type` must be provided when passing an in-memory buffer"
                )
            src = f"data:{mime_type};base64,{b64_str}"

        elif isinstance(file, (str, Path, PurePath)):
            with open(file, "rb") as img_file:
                b64_str = base64.b64encode(img_file.read()).decode("utf-8")
                if mime_type is None:
                    mime_type = mimetypes.guess_type(file)[0]
                src = f"data:{mime_type};base64,{b64_str}"

    card_class_map = {
        "all": "card-img",
        "top": "card-img-top",
        "bottom": "card-img-bottom",
    }

    image = tags.img(
        {
            "src": src,
            "class": "img-fluid",
            "style": css(
                height=as_css_unit(height),
                width=as_css_unit(width),
            ),
        },
        {"class": card_class_map.get(border_radius, None)},
        *args,
        class_=class_,
        **kwargs,
    )

    if fill:
        image = as_fill_item(image)

    if href is not None:
        image = as_fillable_container(
            as_fill_item(
                tags.a(
                    image,
                    href=href,
                )
            )
        )

    if container:
        return container(image)
    else:
        return CardItem(image)



import datetime

from shiny.express import input, render, ui

with ui.card(id="card"):
    ui.input_slider("val", "slider", 0, 100, 50)
    "Text outside of render express call"
    ui.tags.br()
    f"Rendered time: {str(datetime.datetime.now())}"

    @render.express
    def render_express():
        "Text inside of render express call"
        ui.tags.br()
        "Dynamic slider value: "
        input.val()
        ui.tags.br()
        f"Rendered time: {str(datetime.datetime.now())}"



from conftest import ShinyAppProc
from controls import OutputTextVerbatim
from playwright.sync_api import Page


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    OutputTextVerbatim(page, "no_output").expect_value("DATA SUMMARY")
    OutputTextVerbatim(page, "no_parens").expect_value("DATA SUMMARY")
    OutputTextVerbatim(page, "to_upper").expect_value("DATA SUMMARY")
    OutputTextVerbatim(page, "to_lower").expect_value("data summary")



from shiny import App, Inputs, ui

app_ui = ui.page_fixed(
    ui.panel_title("Basic Nav Example"),
    ui.navset_tab(
        ui.nav_panel("One", "First tab content."),
        ui.nav_panel("Two", "Second tab content."),
    ),
)


def server(input: Inputs):
    pass


app = App(app_ui, server)



from __future__ import annotations

import ast
import sys
from pathlib import Path
from typing import cast

from htmltools import Tag, TagList

from .._app import App
from .._docstring import no_example
from ..session import Inputs, Outputs, Session, session_context
from ._mock_session import MockSession
from ._recall_context import RecallContextManager
from .expressify_decorator._func_displayhook import _expressify_decorator_function_def
from .expressify_decorator._node_transformers import (
    DisplayFuncsTransformer,
    expressify_decorator_func_name,
)

__all__ = ("wrap_express_app",)


@no_example()
def wrap_express_app(file: Path) -> App:
    """Wrap a Shiny Express mode app into a Shiny `App` object.

    Parameters
    ----------
    file
        The path to the file containing the Shiny express application.

    Returns
    -------
    :
        A :class:`shiny.App` object.
    """
    try:
        with session_context(cast(Session, MockSession())):
            # We tagify here, instead of waiting for the App object to do it when it wraps
            # the UI in a HTMLDocument and calls render() on it. This is because
            # AttributeErrors can be thrown during the tagification process, and we need to
            # catch them here and convert them to a different type of error, because uvicorn
            # specifically catches AttributeErrors and prints an error message that is
            # misleading for Shiny Express. https://github.com/posit-dev/py-shiny/issues/937
            app_ui = run_express(file).tagify()

    except AttributeError as e:
        raise RuntimeError(e) from e

    def express_server(input: Inputs, output: Outputs, session: Session):
        try:
            run_express(file)

        except Exception:
            import traceback

            traceback.print_exception(*sys.exc_info())
            raise

    app = App(app_ui, express_server)

    return app


def run_express(file: Path) -> Tag | TagList:
    with open(file) as f:
        content = f.read()

    tree = ast.parse(content, file)
    tree = DisplayFuncsTransformer().visit(tree)
    tree = ast.fix_missing_locations(tree)

    ui_result: Tag | TagList = TagList()

    def set_result(x: object):
        nonlocal ui_result
        ui_result = cast(Tag, x)

    prev_displayhook = sys.displayhook
    sys.displayhook = set_result

    try:
        reset_top_level_recall_context_manager()
        get_top_level_recall_context_manager().__enter__()

        file_path = str(file.resolve())

        var_context: dict[str, object] = {
            "__file__": file_path,
            expressify_decorator_func_name: _expressify_decorator_function_def,
            "input": InputNotImportedShim(),
        }

        # Execute each top-level node in the AST
        for node in tree.body:
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                exec(
                    compile(ast.Module([node], type_ignores=[]), file_path, "exec"),
                    var_context,
                    var_context,
                )
            else:
                exec(
                    compile(
                        ast.Interactive([node], type_ignores=[]), file_path, "single"
                    ),
                    var_context,
                    var_context,
                )

        # When we called the function to get the top level recall context manager, we didn't
        # store the result in a variable and re-use that variable here. That is intentional,
        # because during the evaluation of the app code,
        # replace_top_level_recall_context_manager() may have been called, which swaps
        # out the context manager, and it's the new one that we need to exit here.
        get_top_level_recall_context_manager().__exit__(None, None, None)

        # If we're running as an Express app but there's also a top-level item named app
        # which is a shiny.App object, the user probably made a mistake.
        if "app" in var_context and isinstance(var_context["app"], App):
            raise RuntimeError(
                "This looks like a Shiny Express app because it imports shiny.express, "
                "but it also looks like a Shiny Core app because it has a variable named "
                "`app` which is a shiny.App object. Remove either the shiny.express import, "
                "or the app=App()."
            )

        return ui_result

    except AttributeError as e:
        # Need to catch AttributeError and convert to a different type of error, because
        # uvicorn specifically catches AttributeErrors and prints an error message that
        # is helpful for normal ASGI apps, but misleading in the case of Shiny Express.
        raise RuntimeError(e) from e

    finally:
        sys.displayhook = prev_displayhook


_top_level_recall_context_manager: RecallContextManager[Tag]


def reset_top_level_recall_context_manager() -> None:
    from .ui._page import page_auto_cm

    global _top_level_recall_context_manager
    _top_level_recall_context_manager = page_auto_cm()


def get_top_level_recall_context_manager() -> RecallContextManager[Tag]:
    return _top_level_recall_context_manager


class InputNotImportedShim:
    # This is a dummy class that is used to provide a helpful error message when the
    # user tries to access `input.x` but forgot to import `input`. If they do that, then
    # it would get the builtin `input` function, and print an unhelpful error message:
    #   RuntimeError: 'builtin_function_or_method' object has no attribute 'x'
    # This class provides an error message that is more helpful.
    def __getattr__(self, name: str):
        raise AttributeError(
            "Tried to access `input`, but it was not imported. "
            "Perhaps you need `from shiny.express import input`?"
        )



from __future__ import annotations

from conftest import ShinyAppProc
from playwright.sync_api import Page, expect


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # The purpose of this test is to make sure that the data grid can work on Pandas
    # data frames that use an index that is not simply 0-based integers.

    row1 = page.locator("#grid tbody tr:nth-child(1)")
    row3 = page.locator("#grid tbody tr:nth-child(3)")
    result_loc = page.locator("#detail tbody tr:nth-child(1) td:nth-child(1)")
    debug_loc = page.locator("#debug")

    expect(row3).to_be_visible()
    expect(row3.locator("td:nth-child(1)")).to_have_text("three")
    expect(debug_loc).to_have_text("()")

    expect(result_loc).not_to_be_attached()
    row3.click()
    expect(result_loc).to_have_text("three")
    expect(debug_loc).to_have_text("(2,)")

    # Ensure that keys are in sorted order, not the order in which they were selected
    row1.click()
    expect(debug_loc).to_have_text("(0, 2)")



from pathlib import Path

import pandas as pd
import seaborn as sns

from shiny import reactive
from shiny.express import input, render, ui

sns.set_theme(style="white")
df = pd.read_csv(Path(__file__).parent / "penguins.csv", na_values="NA")
species = ["Adelie", "Gentoo", "Chinstrap"]

ui.page_opts(fillable=True)


def count_species(df, species):
    return df[df["Species"] == species].shape[0]


with ui.sidebar():
    ui.input_slider("mass", "Mass", 2000, 6000, 3400)
    ui.input_checkbox_group("species", "Filter by species", species, selected=species)


@reactive.Calc
def filtered_df() -> pd.DataFrame:
    filt_df = df[df["Species"].isin(input.species())]
    filt_df = filt_df.loc[filt_df["Body Mass (g)"] > input.mass()]
    return filt_df


with ui.layout_columns():
    with ui.value_box(theme="primary"):
        "Adelie"

        @render.text
        def adelie_count():
            return count_species(filtered_df(), "Adelie")

    with ui.value_box(theme="primary"):
        "Gentoo"

        @render.text
        def gentoo_count():
            return count_species(filtered_df(), "Gentoo")

    with ui.value_box(theme="primary"):
        "Chinstrap"

        @render.text
        def chinstrap_count():
            return count_species(filtered_df(), "Chinstrap")


with ui.layout_columns():
    with ui.card():
        ui.card_header("Summary statistics")

        @render.data_frame
        def summary_statistics():
            display_df = filtered_df()[
                [
                    "Species",
                    "Island",
                    "Bill Length (mm)",
                    "Bill Depth (mm)",
                    "Body Mass (g)",
                ]
            ]
            return render.DataGrid(display_df, filters=True)

    with ui.card():
        ui.card_header("Penguin bills")

        @render.plot
        def length_depth():
            return sns.scatterplot(
                data=filtered_df(),
                x="Bill Length (mm)",
                y="Bill Depth (mm)",
                hue="Species",
            )



from shiny import ui

_s = ui.sidebar("Sidebar!")
_m = "Body"
_ps = ui.panel_sidebar("Panel Sidebar!")
_pm = ui.panel_main("Panel Main!", TestAttr=True)


def test_panel_main_and_panel_sidebar():
    ps = ui.layout_sidebar(_ps, _pm)
    ps_str = str(ps.tagify())
    assert "TestAttr" in ps_str
    assert "Panel Sidebar!" in ps_str
    assert "Panel Main!" in ps_str

    # OK
    ui.layout_sidebar(_s)
    ui.layout_sidebar(_s, None)

    try:
        ui.layout_sidebar(_s, _s)
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "multiple `sidebar()` objects" in str(e)

    try:
        ui.layout_sidebar(None, _ps)  # pyright: ignore[reportArgumentType]
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "not being supplied with a `sidebar()` object." in str(e)

    try:
        ui.layout_sidebar(_s, _pm)
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "is not being used with `panel_sidebar()`" in str(e)

    try:
        ui.layout_sidebar(_ps, None, _pm)
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "not being supplied as the second argument" in str(e)
    try:
        ui.layout_sidebar(_ps, _pm, None, "42")
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "Unexpected extra legacy `*args`" in str(e)



# # Within file flags to ignore unused imports
# flake8: noqa: F401
# pyright: reportUnusedImport=false

__all__ = (
    "Concatenate",
    "ParamSpec",
    "TypeGuard",
    "NotRequired",
    "TypedDict",
    "assert_type",
)


import sys

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec, TypeGuard
else:
    from typing_extensions import Concatenate, ParamSpec, TypeGuard

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, Self, TypedDict, assert_type
else:
    from typing_extensions import NotRequired, Self, TypedDict, assert_type


# The only purpose of the following line is so that pyright will put all of the
# conditional imports into the .pyi file when generating type stubs. Without this line,
# pyright will not include the above imports in the generated .pyi file, and it will
# result in a lot of red squiggles in user code.
_: 'Concatenate[str, ParamSpec("P")] | ParamSpec | TypeGuard | NotRequired | TypedDict | assert_type | Self'  # type:ignore



import pandas as pd

from shiny import App, Inputs, Outputs, Session, render, ui

df = pd.DataFrame(
    dict(
        id=["one", "two", "three"],
        fname=["Alice", "Bob", "Charlie"],
        lname=["Smith", "Jones", "Brown"],
    )
).set_index(  # type: ignore
    "id",
    drop=False,
)

app_ui = ui.page_fluid(
    ui.p("Select rows in the grid, make sure the selected rows appear below."),
    ui.output_data_frame("grid"),
    ui.output_table("detail"),
    ui.output_text_verbatim("debug"),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.data_frame
    def grid():
        return render.DataGrid(
            df,
            row_selection_mode="multiple",
            height=350,
        )

    @render.table
    def detail():
        if (
            input.grid_selected_rows() is not None
            and len(input.grid_selected_rows()) > 0
        ):
            # "split", "records", "index", "columns", "values", "table"
            return df.iloc[list(input.grid_selected_rows())]

    @render.text
    def debug():
        return input.grid_selected_rows()


app = App(app_ui, server, debug=True)



from datetime import date

from shiny.express import ui

ui.input_date_range("daterange1", "Date range:", start="2001-01-01", end="2010-12-31")
# Default start and end is the current date in the client's time zone
ui.input_date_range("daterange2", "Date range:")
# start and end are always specified in yyyy-mm-dd, even if the display
# format is different
ui.input_date_range(
    "daterange3",
    "Date range:",
    start="2001-01-01",
    end="2010-12-31",
    min="2001-01-01",
    max="2012-12-21",
    format="mm/dd/yy",
    separator=" - ",
)
# Pass in Date objects
ui.input_date_range(
    "daterange4", "Date range:", start=date(2001, 1, 1), end=date(2010, 12, 31)
)
# Use different language and different first day of week
ui.input_date_range("daterange5", "Date range:", language="de", weekstart=1)
# Start with decade view instead of default month view
ui.input_date_range("daterange6", "Date range:", startview="decade")



from conftest import ShinyAppProc
from controls import OutputTextVerbatim
from playwright.sync_api import Page


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    OutputTextVerbatim(page, "all_txt").expect_value(
        "('a-3-flush', 'bx-3-first-flush', 'by-3-first-flush', 'bx-3-second-flush', "
        "'by-3-second-flush', 'c-3-flush', 'a-3-flushed', 'bx-3-first-flushed', "
        "'by-3-first-flushed', 'bx-3-second-flushed', 'by-3-second-flushed', "
        "'c-3-flushed')"
    )
    OutputTextVerbatim(page, "flush_txt").expect_value(
        "('a-3-flush', 'bx-3-first-flush', 'by-3-first-flush', 'bx-3-second-flush', "
        "'by-3-second-flush', 'c-3-flush')"
    )
    OutputTextVerbatim(page, "flushed_txt").expect_value(
        "('a-3-flushed', 'bx-3-first-flushed', 'by-3-first-flushed', "
        "'bx-3-second-flushed', 'by-3-second-flushed', 'c-3-flushed')"
    )

    # Verify `on_ended` callbacks are called in the correct order (and cancelled)
    local_app.close()

    # Wait up to 3 seconds for the app to close and print the logs. (Should be ~ instant)
    local_app.stdout.wait_for(lambda x: "test4" in x, 3)
    stdout = str(local_app.stdout)
    out_indexes = [
        stdout.index("session ended - sync - test1"),
        stdout.index("session ended - async - test2"),
        stdout.index("session ended - async - test3"),
        stdout.index("session ended - sync - test4"),
    ]
    for i in range(len(out_indexes)):
        index = out_indexes[i]
        assert index >= 0
        # Make sure they are ordered correctly
        if i > 0:
            prev_index = out_indexes[i - 1]
            assert index > prev_index



from __future__ import annotations

from colors import bg_color, fg_color
from conftest import ShinyAppProc
from controls import Sidebar, _expect_class_value
from playwright.sync_api import Page, expect


def test_colors_are_rgb() -> None:
    assert bg_color.startswith("rgb(")
    assert fg_color.startswith("rgb(")


def test_sidebar_bg_colors(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    for i in range(1, 5):
        content = page.locator(f"#m{i}")
        sidebar = page.locator(f"#s{i}")

        main_layout = sidebar.locator("..")

        open_val = "always" if i <= 2 else "desktop"
        position_val = "left" if i % 2 == 1 else "right"

        expect(main_layout).to_have_attribute("data-bslib-sidebar-open", open_val)

        _expect_class_value(main_layout, "sidebar-right", position_val == "right")

        expect(content).to_have_text(f"Main content - {i}")
        expect(sidebar).to_have_text(f"Sidebar content - {i}")

        # Only works if css file is loaded
        expect(sidebar).to_have_css("background-color", bg_color)
        expect(sidebar).to_have_css("color", fg_color)

    s1 = Sidebar(page, "s1")
    s1.expect_position("left")
    s2 = Sidebar(page, "s2")
    s2.expect_position("right")



from typing import Callable

import pandas as pd
from plots import plot_auc_curve, plot_precision_recall_curve, plot_score_distribution

from shiny import Inputs, Outputs, Session, module, render, ui


@module.ui
def training_ui():
    return ui.nav_panel(
        "Training Dashboard",
        ui.layout_columns(
            ui.card(
                ui.card_header("Model Metrics"),
                ui.output_plot("metric"),
                ui.input_select(
                    "metric",
                    "Metric",
                    choices=["ROC Curve", "Precision-Recall"],
                ),
            ),
            ui.card(
                ui.card_header("Training Scores"),
                ui.output_plot("score_dist"),
            ),
        ),
    )


@module.server
def training_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    df: Callable[[], pd.DataFrame],
):
    @render.plot
    def score_dist():
        return plot_score_distribution(df())

    @render.plot
    def metric():
        if input.metric() == "ROC Curve":
            return plot_auc_curve(df(), "is_electronics", "training_score")
        else:
            return plot_precision_recall_curve(df(), "is_electronics", "training_score")


@module.ui
def data_view_ui():
    return ui.nav_panel(
        "View Data",
        ui.layout_columns(
            ui.value_box(
                title="Row count",
                value=ui.output_text("row_count"),
                theme="primary",
            ),
            ui.value_box(
                title="Mean score",
                value=ui.output_text("mean_score"),
                theme="bg-green",
            ),
            gap="20px",
        ),
        ui.layout_columns(
            ui.card(ui.output_data_frame("data")),
            style="margin-top: 20px;",
        ),
    )


@module.server
def data_view_server(
    input: Inputs, output: Outputs, session: Session, df: Callable[[], pd.DataFrame]
):
    @render.text
    def row_count():
        return df().shape[0]

    @render.text
    def mean_score():
        return round(df()["training_score"].mean(), 2)

    @render.data_frame
    def data():
        return df()



from conftest import ShinyAppProc
from controls import OutputTextVerbatim
from playwright.sync_api import Page
from playwright.sync_api import expect as playwright_expect


def test_express_page_fluid(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    txt = OutputTextVerbatim(page, "visible")
    txt.expect_value("40")

    playwright_expect(page.locator("#visible")).to_have_count(1)
    playwright_expect(page.locator("#hidden")).to_have_count(0)



from typing import List

from shiny import App, Inputs, Outputs, Session, reactive, ui
from shiny.types import NavSetArg


def nav_controls(prefix: str) -> List[NavSetArg]:
    return [
        ui.nav_panel("a", prefix + ": tab a content"),
        ui.nav_panel("b", prefix + ": tab b content"),
        ui.nav_panel("c", prefix + ": tab c content"),
        ui.nav_spacer(),
        ui.nav_menu(
            "Links",
            ui.nav_control(
                ui.a(
                    "Shiny",
                    href="https://shiny.posit.co/py/",
                    target="_blank",
                )
            ),
            "----",
            "Plain text",
            "----",
            ui.nav_control(
                ui.a(
                    "Posit",
                    href="https://posit.co",
                    target="_blank",
                )
            ),
            align="right",
        ),
    ]


app_ui = ui.page_navbar(
    *nav_controls("page_navbar"),
    title="page_navbar()",
    id="navbar_id",
    footer=ui.div(
        {"style": "width:80%;margin: 0 auto"},
        ui.tags.style(
            """
            h4 {
                margin-top: 3em;
            }
            """
        ),
        ui.h4("navset_tab()"),
        ui.navset_tab(*nav_controls("navset_tab()")),
        ui.h4("navset_pill()"),
        ui.navset_pill(*nav_controls("navset_pill()")),
        ui.h4("navset_underline()"),
        ui.navset_underline(*nav_controls("navset_underline()")),
        ui.h4("navset_card_tab()"),
        ui.navset_card_tab(*nav_controls("navset_card_tab()")),
        ui.h4("navset_card_pill()"),
        ui.navset_card_pill(*nav_controls("navset_card_pill()")),
        ui.h4("navset_card_underline()"),
        ui.navset_card_underline(*nav_controls("navset_card_underline()")),
        ui.h4("navset_pill_list()"),
        ui.navset_pill_list(*nav_controls("navset_pill_list()")),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        print("Current navbar page: ", input.navbar_id())


app = App(app_ui, server)



from __future__ import annotations

import random
import warnings
from typing import TYPE_CHECKING, Literal, Optional, cast

from htmltools import (
    HTML,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    css,
    div,
    tags,
)

from .._deprecated import warn_deprecated
from .._docstring import add_example, no_example
from .._namespaces import resolve_id_or_none
from ..session import require_active_session
from ._card import CardItem
from ._html_deps_shinyverse import components_dependency
from ._tag import consolidate_attrs, trinary
from ._utils import css_no_sub
from .css import CssUnit, as_css_padding, as_css_unit
from .fill import as_fill_item, as_fillable_container

if TYPE_CHECKING:
    from .. import Session

__all__ = (
    "Sidebar",
    "sidebar",
    "layout_sidebar",
    "update_sidebar",
    # Legacy
    "panel_sidebar",
    "panel_main",
)


@no_example()
class Sidebar:
    """
    A sidebar object

    Class returned from :func:`~shiny.ui.sidebar`. Please do not use this
    class directly. Instead, supply the :func:`~shiny.ui.sidebar` object to
    :func:`~shiny.ui.layout_sidebar`.

    Attributes
    ----------
    tag
        The :class:`~htmltools.Tag` object that represents the sidebar.
    collapse_tag
        The :class:`~htmltools.Tag` object that represents the collapse button.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar (open or collapsed).
    width
        A valid CSS unit used for the width of the sidebar.
    max_height_mobile
        The maximum height of the horizontal sidebar when viewed on mobile devices.
        The default is `250px` unless the sidebar is included in a
        :func:`~shiny.ui.layout_sidebar` with a specified height, in
        which case the default is to take up no more than 50% of the layout container.
    color_fg
        A foreground color.
    color_bg
        A background color.

    Parameters
    ----------
    tag
        The :class:`~htmltools.Tag` object that represents the sidebar.
    collapse_tag
        The :class:`~htmltools.Tag` object that represents the collapse button.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar (open or collapsed).
    width
        A valid CSS unit used for the width of the sidebar.
    max_height_mobile
        The maximum height of the horizontal sidebar when viewed on mobile devices.
        The default is `250px` unless the sidebar is included in a
        :func:`~shiny.ui.layout_sidebar` with a specified height, in
        which case the default is to take up no more than 50% of the layout container.
    color_fg
        A foreground color.
    color_bg
        A background color.
    """

    def __init__(
        self,
        tag: Tag,
        collapse_tag: Optional[Tag],
        position: Literal["left", "right"],
        open: Literal["desktop", "open", "closed", "always"],
        width: CssUnit,
        max_height_mobile: Optional[str | float],
        color_fg: Optional[str],
        color_bg: Optional[str],
    ):
        self.tag = tag
        self.collapse_tag = collapse_tag
        self.position = position
        self.open = open
        self.width = width
        self.max_height_mobile = max_height_mobile
        self.color_fg = color_fg
        self.color_bg = color_bg

    # The `Sidebar` class should use it's fields, not this method
    def tagify(self) -> Tag:
        """
        Not implemented
        """
        # Similar to `NavMenu.tagify()`
        raise NotImplementedError(
            "`Sidebar` objects must be handled by `layout_sidebar(*args)`."
        )


@add_example()
def sidebar(
    *args: TagChild | TagAttrs,
    width: CssUnit = 250,
    position: Literal["left", "right"] = "left",
    open: Literal["desktop", "open", "closed", "always"] = "desktop",
    id: Optional[str] = None,
    title: TagChild | str = None,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    class_: Optional[str] = None,  # TODO-future; Consider using `**kwargs` instead
    max_height_mobile: Optional[str | float] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
) -> Sidebar:
    # See [this article](https://rstudio.github.io/bslib/articles/sidebars.html)
    #   to learn more.
    # TODO-future; If color contrast is implemented. Docs for `bg` and `fg`:
    #     If only one of either is provided, an
    #     accessible contrasting color is provided for the opposite color, e.g. setting
    #     `bg` chooses an appropriate `fg` color.

    """
    Sidebar element

    Create a collapsing sidebar layout by providing a `sidebar()` object to the
    `sidebar=` argument of:

    * :func:`~shiny.ui.layout_sidebar`
      * Creates a sidebar layout component which can be dropped inside any Shiny UI page method (e.g. :func:`~shiny.ui.page_fillable`) or :func:`~shiny.ui.card` context.
    * :func:`~shiny.ui.navset_bar`, :func:`~shiny.ui.navset_card_tab`, and :func:`~shiny.ui.navset_card_pill`
      * Creates a multi page/tab UI with a singular `sidebar()` (which is
        shown on every page/tab).

    Parameters
    ----------
    *args
        Contents to the sidebar. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar.

        * `"desktop"`: the sidebar starts open on desktop screen, closed on mobile
        * `"open"` or `True`: the sidebar starts open
        * `"closed"` or `False`: the sidebar starts closed
        * `"always"` or `None`: the sidebar is always open and cannot be closed

        In :func:`~shiny.ui.update_sidebar`, `open` indicates the desired state of the
        sidebar. Note that :func:`~shiny.ui.update_sidebar` can only open or close the
        sidebar, so it does not support the `"desktop"` and `"always"` options.
    id
        A character string. Required if wanting to re-actively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :class:`~htmltools.Tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    bg,fg
        A background or foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        A CSS length unit (passed through :func:`~shiny.ui.css.as_css_unit`) defining
        the maximum height of the horizontal sidebar when viewed on mobile devices. Only
        applies to always-open sidebars that use `open = "always"`, where by default the
        sidebar container is placed below the main content container on mobile devices.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.

    Returns
    -------
    :
        A :class:`~shiny.ui.Sidebar` object.

    See Also
    --------
    * :func:`~shiny.ui.layout_sidebar`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    """
    # TODO-future; validate `open`, bg, fg, class_, max_height_mobile

    if max_height_mobile is not None and open != "always":
        warnings.warn(
            "The `shiny.ui.sidebar(max_height_mobile=)` argument only applies to the sidebar when `open = 'always'`. The `max_height_mobile` argument will be ignored.",
            # `stacklevel=2`: Refers to the caller of `sidebar()`
            stacklevel=2,
        )
        max_height_mobile = None

    if id is None and open != "always":
        # but always provide id when collapsible for accessibility reasons
        id = f"bslib_sidebar_{random.randint(1000, 10000)}"
    resolved_id = resolve_id_or_none(id)

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    if isinstance(title, (str, int, float)):
        title = tags.header(str(title), class_="sidebar-title")

    collapse_tag = None
    if open != "always":
        collapse_tag = tags.button(
            _collapse_icon(),
            class_="collapse-toggle",
            type="button",
            title="Toggle sidebar",
            aria_expanded=trinary(open in ["open", "desktop"]),
            aria_controls=resolved_id,
        )

    tag = tags.aside(
        div(
            title,
            {
                "class": "sidebar-content bslib-gap-spacing",
                "style": css(
                    gap=as_css_unit(gap),
                    padding=as_css_padding(padding),
                ),
            },
            *args,
        ),
        {"class": "bslib-sidebar-input"} if resolved_id is not None else None,
        {
            "class": "sidebar",
            "hidden": True if (open in ["closed", "desktop"]) else None,
        },
        id=resolved_id,
        class_=class_,
    )

    return Sidebar(
        tag=tag,
        collapse_tag=collapse_tag,
        position=position,
        open=open,
        width=width,
        max_height_mobile=max_height_mobile,
        color_fg=fg,
        color_bg=bg,
    )


@add_example()
def layout_sidebar(
    sidebar: Sidebar,
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    fill: bool = True,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    border: Optional[bool] = None,
    border_radius: Optional[bool] = None,
    border_color: Optional[str] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Sidebar layout

    Create a sidebar layout component which can be dropped inside any Shiny UI page
    method (e.g. :func:`~shiny.ui.page_fillable`) or
    :func:`~shiny.ui.card` context.

    Parameters
    ----------
    *args
        One argument needs to be of class :class:`~shiny.ui.Sidebar` object created by
        :func:`~shiny.ui.sidebar`. The remaining arguments will contain the contents to
        the main content area. Or tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    fillable
        Whether or not the main content area should be wrapped in a fillable container.
        See :func:`~shiny.ui.fill.as_fillable_container` for details.
    fill
        Whether or not the sidebar layout should be wrapped in a fillable container. See
        :func:`~shiny.ui.fill.as_fill_item` for details.
    bg,fg
        A background or foreground color.
    border
        Whether or not to show a border around the sidebar layout.
    border_radius
        Whether or not to round the corners of the sidebar layout.
    border_color
        A border color.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`. This value will only be used if `fillable` is `True`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.
    height
        Any valid CSS unit to use for the height.

    Returns
    -------
    :
        A :class:`~htmltools.Tag` object.

    See Also
    --------
    * :func:`~shiny.ui.sidebar`
    """

    sidebar, args = _get_layout_sidebar_sidebar(sidebar, args)

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    attrs, children = consolidate_attrs(*args, **kwargs)
    # TODO-future: >= 2023-11-01); Once `panel_main()` is removed, we can remove this loop
    for child in children:
        if isinstance(child, DeprecatedPanelMain):
            attrs = consolidate_attrs(attrs, child.attrs)[0]
            # child.children will be handled when tagified

    main = div(
        {
            "class": f"main{' bslib-gap-spacing' if fillable else ''}",
            ""
            "style": css(
                gap=as_css_unit(gap),
                padding=as_css_padding(padding),
            ),
        },
        attrs,
        *children,
    )
    if fillable:
        main = as_fillable_container(main)

    max_height_mobile = sidebar.max_height_mobile or (
        "250px" if height is None else "50%"
    )

    res = div(
        {"class": "bslib-sidebar-layout bslib-mb-spacing"},
        {"class": "sidebar-right"} if sidebar.position == "right" else None,
        {"class": "sidebar-collapsed"} if sidebar.open == "closed" else None,
        main,
        sidebar.tag,
        sidebar.collapse_tag,
        components_dependency(),
        _sidebar_init_js(),
        data_bslib_sidebar_init="true",
        data_bslib_sidebar_open=sidebar.open,
        data_bslib_sidebar_border=trinary(border),
        data_bslib_sidebar_border_radius=trinary(border_radius),
        style=css_no_sub(
            **{
                "--_sidebar-width": as_css_unit(sidebar.width),
                "--_sidebar-bg": sidebar.color_bg,
                "--_sidebar-fg": sidebar.color_fg,
                "--_main-fg": fg,
                "--_main-bg": bg,
                "--bs-card-border-color": border_color,
                "height": as_css_unit(height),
                "--_mobile-max-height": as_css_unit(max_height_mobile),
            },
        ),
    )
    if fill:
        res = as_fill_item(res)

    return CardItem(res)


def _get_layout_sidebar_sidebar(
    sidebar: Sidebar,
    args: tuple[TagChild | TagAttrs, ...],
) -> tuple[Sidebar, tuple[TagChild | TagAttrs, ...]]:
    updated_args: list[TagChild | TagAttrs] = []
    original_args = tuple(args)

    # sidebar: Sidebar | None = None
    sidebar_orig_arg: Sidebar | DeprecatedPanelSidebar = sidebar

    if isinstance(sidebar, DeprecatedPanelSidebar):
        sidebar = sidebar.sidebar

    if not isinstance(sidebar, Sidebar):
        raise ValueError(
            "`layout_sidebar()` is not being supplied with a `sidebar()` object. Please supply a `sidebar()` object to `layout_sidebar(sidebar)`."
        )

    # Use `original_args` here so `updated_args` can be safely altered in place
    for i, arg in zip(range(len(original_args)), original_args):
        if isinstance(arg, DeprecatedPanelSidebar):
            raise ValueError(
                "`panel_sidebar()` is not being used as the first argument to `layout_sidebar(sidebar,)`. `panel_sidebar()` has been deprecated and will go away in a future version of Shiny. Please supply `panel_sidebar()` arguments directly to `args` in `layout_sidebar(sidebar)` and use `sidebar()` instead of `panel_sidebar()`."
            )
        elif isinstance(arg, Sidebar):
            raise ValueError(
                "`layout_sidebar()` is being supplied with multiple `sidebar()` objects. Please supply only one `sidebar()` object to `layout_sidebar()`."
            )

        elif isinstance(arg, DeprecatedPanelMain):
            if i != 0:
                raise ValueError(
                    "`panel_main()` is not being supplied as the second argument to `layout_sidebar()`. `panel_main()`/`panel_sidebar()` have been deprecated and will go away in a future version of Shiny. Please supply `panel_main()` arguments directly to `args` in `layout_sidebar(sidebar, *args)` and use `sidebar()` instead of `panel_sidebar()`."
                )
            if not isinstance(sidebar_orig_arg, DeprecatedPanelSidebar):
                raise ValueError(
                    "`panel_main()` is not being used with `panel_sidebar()`. `panel_main()`/`panel_sidebar()` have been deprecated and will go away in a future version of Shiny. Please supply `panel_main()` arguments directly to `args` in `layout_sidebar(sidebar, *args)` and use `sidebar()` instead of `panel_sidebar()`."
                )

            if len(args) > 2:
                raise ValueError(
                    "Unexpected extra legacy `*args` have been supplied to `layout_sidebar()` in addition to `panel_main()` or `panel_sidebar()`. `panel_main()` has been deprecated and will go away in a future version of Shiny. Please supply `panel_main()` arguments directly to `args` in `layout_sidebar(sidebar, *args)` and use `sidebar()` instead of `panel_sidebar()`."
                )
            # Notes for this point in the code:
            # * We are working with args[0], a `DeprecatedPanelMain`; sidebar was originally a `DeprecatedPanelSidebar`
            # * len(args) == 1 or 2

            # Handle legacy `layout_sidebar(sidebar, main, position=)` value
            if len(args) == 2:
                arg1 = args[1]
                if not (arg1 == "left" or arg1 == "right"):
                    raise ValueError(
                        "layout_sidebar(*args) contains non-valid legacy values. Please use `sidebar()` instead of `panel_sidebar()` and supply any `panel_main()` arguments directly to `args` in `layout_sidebar(sidebar, *args)`."
                    )
                # We know `sidebar_orig_arg` is a `DeprecatedPanelSidebar` here
                sidebar.position = cast(  # pyright: ignore[reportOptionalMemberAccess]
                    Literal["left", "right"],
                    arg1,
                )

            # Only keep panel_main content
            updated_args = [arg]

            # Cases have been covered, quit loop
            break

            # Extract `DeprecatedPanelMain` attrs and children in followup for loop
        else:
            # Keep the arg!
            updated_args.append(arg)

    return (sidebar, tuple(updated_args))


@add_example()
def update_sidebar(
    id: str,
    *,
    show: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Update a sidebar's visibility.

    Set a :func:`~shiny.ui.sidebar` state during an active Shiny user session.

    Parameters
    ----------
    id
        The `id` of the :func:`~shiny.ui.sidebar` to toggle.
    show
        The desired visible state of the sidebar, where `True` opens the sidebar and `False` closes the sidebar (if not already in that state).
    session
        A Shiny session object (the default should almost always be used).

    See Also
    --------
    * :func:`~shiny.ui.sidebar`
    * :func:`~shiny.ui.layout_sidebar`
    """
    session = require_active_session(session)

    # method: Literal["toggle", "open", "close"]
    # if open is None or open == "toggle":
    #     method = "toggle"
    # elif open is True or open == "open":
    #     method = "open"
    # elif open is False or open == "closed":
    #     method = "close"
    # else:
    #     if open == "always" or open == "desktop":
    #         raise ValueError(
    #             f"`open = '{open}'` is not supported by `update_sidebar()`"
    #         )
    #     raise ValueError(
    #         "open must be NULL (or 'toggle'), TRUE (or 'open'), or FALSE (or 'closed')"
    #     )
    if show is not None:
        method = "open" if bool(show) else "close"

        def callback() -> None:
            session.send_input_message(id, {"method": method})

        session.on_flush(callback, once=True)


def _collapse_icon() -> TagChild:
    return HTML(
        '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-chevron-left collapse-icon" style="fill:currentColor;" aria-hidden="true" role="img" ><path fill-rule="evenodd" d="M11.354 1.646a.5.5 0 0 1 0 .708L5.707 8l5.647 5.646a.5.5 0 0 1-.708.708l-6-6a.5.5 0 0 1 0-.708l6-6a.5.5 0 0 1 .708 0z"></path></svg>'
    )


def _sidebar_init_js() -> Tag:
    # Note: if we want to avoid inline `<script>` tags in the future for
    # initialization code, we might be able to do so by turning the sidebar layout
    # container into a web component
    return tags.script(
        {"data-bslib-sidebar-init": True},
        "bslib.Sidebar.initCollapsibleAll()",
    )


######################


# Deprecated 2023-06-13
# Includes: DeprecatedPanelSidebar
@no_example()
def panel_sidebar(
    *args: TagChild | TagAttrs,
    width: int = 4,
    **kwargs: TagAttrValue,
) -> DeprecatedPanelSidebar:
    """Deprecated. Please use :func:`~shiny.ui.sidebar` instead."""
    # TODO-future: >= 2024-01-01; Add deprecation message below
    # Plan of action:
    # * No deprecation messages today (2023-10-11), and existing code _just works_.
    # * Change all examples to use the new API.
    # * In, say, 6 months, start emitting messages for code that uses the old API.

    # warn_deprecated("Please use `sidebar()` instead of `panel_sidebar()`. `panel_sidebar()` will go away in a future version of Shiny.")
    return DeprecatedPanelSidebar(
        *args,
        width=width,
        **kwargs,
    )


# Deprecated 2023-06-13
# Includes: DeprecatedPanelMain
@no_example()
def panel_main(
    *args: TagChild | TagAttrs,
    width: int = 8,
    **kwargs: TagAttrValue,
) -> DeprecatedPanelMain:
    """Deprecated. Please supply the `*args` of :func:`~shiny.ui.panel_main` directly to :func:`~shiny.ui.layout_sidebar`."""
    # TODO-future: >= 2023-11-01; Add deprecation message below
    # warn_deprecated("Please use `layout_sidebar(*args)` instead of `panel_main()`. `panel_main()` will go away in a future version of Shiny.")

    # warn if keys are being ignored
    attrs, children = consolidate_attrs(*args, **kwargs)
    if len(attrs) > 0:
        return DeprecatedPanelMain(attrs=attrs, children=children)
        warn_deprecated(
            "`*args: TagAttrs` or `**kwargs: TagAttrValue` values supplied to `panel_main()` are being ignored. Please supply them directly to `layout_sidebar()`."
        )

    return DeprecatedPanelMain(attrs={}, children=children)


# Deprecated 2023-06-13


# This class should be removed when `panel_sidebar()` is removed
class DeprecatedPanelSidebar(
    # While it doesn't seem right to inherit from `Sidebar`, it's the easiest way to
    # make sure `layout_sidebar(sidebar: Sidebar)` works without mucking up the
    # function signature.
    Sidebar
):
    """
    [Deprecated] Sidebar panel

    Class returned from :func:`~shiny.ui.panel_sidebar`. Please do not
    use this class and instead supply your content to
    :func:`~shiny.ui.layout_sidebar` directly.

    Parameters
    ----------
    *args
        Contents to the sidebar. Or tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    width
        An integeger between 1 and 12, inclusive, that determines the width of the
        sidebar. The default is 4.
    **kwargs
        Tag attributes that are supplied to the resolved :class:`~htmltools.Tag` object.

    Attributes
    ----------
    sidebar
        A output from :func:`~shiny.ui.sidebar`.

    See Also
    --------
    * :func:`~shiny.ui.layout_sidebar`
    * :func:`~shiny.ui.sidebar`
    """

    # Store `attrs` for `layout_sidebar()` to retrieve
    sidebar: Sidebar

    def __init__(
        self, *args: TagChild | TagAttrs, width: int = 4, **kwargs: TagAttrValue
    ) -> None:
        self.sidebar = sidebar(
            *args,
            width=f"{int(width / 12 * 100)}%",
            open="always",
            **kwargs,  # pyright: ignore[reportArgumentType]
        )

    # Hopefully this is never used. But wanted to try to be safe
    def tagify(self) -> Tag:
        """
        Tagify the `self.sidebar.tag` and return the result in a TagList
        """
        return self.sidebar.tag.tagify()


# This class should be removed when `panel_main()` is removed
# Must be `Tagifiable`, so it can fit as a type `TagChild`
class DeprecatedPanelMain:
    """
    [Deprecated] Main panel

    Class returned from :func:`~shiny.ui.panel_main`. Please do not use
    this class and instead supply your content to
    :func:`~shiny.ui.layout_sidebar` directly.


    Parameters
    ----------
    attrs
        Attributes to apply to the parent tag of the children.
    children
        Children UI Elements to render inside the parent tag.

    Attributes
    ----------
    attrs
        Attributes to apply to the parent tag of the children.
    children
        Children UI Elements to render inside the parent tag.

    See Also
    --------
    * :func:`~shiny.ui.layout_sidebar`
    * :func:`~shiny.ui.sidebar`
    """

    # Store `attrs` for `layout_sidebar()` to retrieve
    attrs: TagAttrs
    # Return `children` in `layout_sidebar()` via `.tagify()` method
    children: list[TagChild]

    def __init__(self, *, attrs: TagAttrs, children: list[TagChild]) -> None:
        self.attrs = attrs
        self.children = children

    def tagify(self) -> TagList:
        """
        Tagify the `children` and return the result in a TagList
        """
        return TagList(self.children).tagify()



from pathlib import Path

import pandas as pd
import seaborn as sns

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

sns.set_theme(style="white")
df = pd.read_csv(Path(__file__).parent / "penguins.csv", na_values="NA")
species = ["Adelie", "Gentoo", "Chinstrap"]


def make_value_box(penguin):
    return ui.value_box(
        title=penguin, value=ui.output_text(f"{penguin}_count".lower()), theme="primary"
    )


app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_slider(
            "mass",
            "Mass",
            2000,
            6000,
            3400,
        ),
        ui.input_checkbox_group(
            "species", "Filter by species", species, selected=species
        ),
    ),
    ui.layout_columns(
        *[make_value_box(penguin) for penguin in species],
    ),
    ui.layout_columns(
        ui.card(
            ui.card_header("Summary statistics"),
            ui.output_data_frame("summary_statistics"),
        ),
        ui.card(
            ui.card_header("Penguin bills"),
            ui.output_plot("length_depth"),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Calc
    def filtered_df() -> pd.DataFrame:
        filt_df = df[df["Species"].isin(input.species())]
        filt_df = filt_df.loc[filt_df["Body Mass (g)"] > input.mass()]
        return filt_df

    @render.text
    def adelie_count():
        return count_species(filtered_df(), "Adelie")

    @render.text
    def chinstrap_count():
        return count_species(filtered_df(), "Chinstrap")

    @render.text
    def gentoo_count():
        return count_species(filtered_df(), "Gentoo")

    @render.plot
    def length_depth():
        return sns.scatterplot(
            data=filtered_df(),
            x="Bill Length (mm)",
            y="Bill Depth (mm)",
            hue="Species",
        )

    @render.data_frame
    def summary_statistics():
        display_df = filtered_df()[
            [
                "Species",
                "Island",
                "Bill Length (mm)",
                "Bill Depth (mm)",
                "Body Mass (g)",
            ]
        ]
        return render.DataGrid(display_df, filters=True)


def count_species(df, species):
    return df[df["Species"] == species].shape[0]


app = App(app_ui, server)



from __future__ import annotations

from typing import Callable

from htmltools import Tag

from ... import ui
from ..._docstring import no_example
from ...types import MISSING, MISSING_TYPE
from .._recall_context import RecallContextManager
from .._run import get_top_level_recall_context_manager

__all__ = ("page_opts",)


def page_auto_cm() -> RecallContextManager[Tag]:
    return RecallContextManager(ui.page_auto)


@no_example()
def page_opts(
    *,
    title: str | MISSING_TYPE = MISSING,
    window_title: str | MISSING_TYPE = MISSING,
    lang: str | MISSING_TYPE = MISSING,
    page_fn: Callable[..., Tag] | None | MISSING_TYPE = MISSING,
    fillable: bool | MISSING_TYPE = MISSING,
    full_width: bool | MISSING_TYPE = MISSING,
) -> None:
    """
    Set page-level options for the current app.

    The arguments to this function get passed to :func:`~shiny.ui.page_auto`.

    Parameters
    ----------
    title
        A title shown on the page.
    window_title
        The browser window title. If no value is provided, this will use the value of
        ``title``.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    fillable
        If there is a top-level sidebar or nav, then the value is passed through to the
        :func:`~shiny.ui.page_sidebar` or :func:`~shiny.ui.page_navbar` function.
        Otherwise, if ``True``, use :func:`~shiny.ui.page_fillable`, where the content
        fills the window; if ``False`` (the default), the value of ``full_width`` will
        determine which page function is used.
    full_width
        This has an effect only if there are no sidebars or top-level navs, and
        ``fillable`` is ``False``. If this is ``False`` (the default), use use
        :func:`~shiny.ui.page_fixed`; if ``True``, use :func:`~shiny.ui.page_fillable`.
    page_fn
        The page function to use. If ``None`` (the default), will automatically choose
        one based on the arguments provided. If not ``None``, this will override all
        heuristics for choosing page functions.
    """
    cm = get_top_level_recall_context_manager()

    if not isinstance(title, MISSING_TYPE):
        cm.kwargs["title"] = title
    if not isinstance(window_title, MISSING_TYPE):
        cm.kwargs["window_title"] = window_title
    if not isinstance(lang, MISSING_TYPE):
        cm.kwargs["lang"] = lang
    if not isinstance(page_fn, MISSING_TYPE):
        cm.kwargs["page_fn"] = page_fn
    if not isinstance(fillable, MISSING_TYPE):
        cm.kwargs["fillable"] = fillable
    if not isinstance(full_width, MISSING_TYPE):
        cm.kwargs["full_width"] = full_width



from pandas import DataFrame
from plotnine import (
    aes,
    geom_abline,
    geom_density,
    geom_line,
    ggplot,
    labs,
    theme_minimal,
)
from sklearn.metrics import auc, precision_recall_curve, roc_curve


def plot_score_distribution(df: DataFrame):
    plot = (
        ggplot(df, aes(x="training_score"))
        + geom_density(fill="blue", alpha=0.3)
        + theme_minimal()
        + labs(title="Model scores", x="Score")
    )
    return plot


def plot_auc_curve(df: DataFrame, true_col: str, pred_col: str):
    fpr, tpr, _ = roc_curve(df[true_col], df[pred_col])
    roc_auc = auc(fpr, tpr)

    roc_df = DataFrame({"fpr": fpr, "tpr": tpr})

    plot = (
        ggplot(roc_df, aes(x="fpr", y="tpr"))
        + geom_line(color="darkorange", size=1.5, show_legend=True, linetype="solid")
        + geom_abline(intercept=0, slope=1, color="navy", linetype="dashed")
        + labs(
            title="Receiver Operating Characteristic (ROC)",
            subtitle=f"AUC: {roc_auc.round(2)}",
            x="False Positive Rate",
            y="True Positive Rate",
        )
        + theme_minimal()
    )

    return plot


def plot_precision_recall_curve(df: DataFrame, true_col: str, pred_col: str):
    precision, recall, _ = precision_recall_curve(df[true_col], df[pred_col])

    pr_df = DataFrame({"precision": precision, "recall": recall})

    plot = (
        ggplot(pr_df, aes(x="recall", y="precision"))
        + geom_line(color="darkorange", size=1.5, show_legend=True, linetype="solid")
        + labs(
            title="Precision-Recall Curve",
            x="Recall",
            y="Precision",
        )
        + theme_minimal()
    )

    return plot



from __future__ import annotations

import typing

from shiny import App, Inputs, Outputs, Session, render, ui

slider_nums: list[int] = []


def slider_row(
    label: str,
    **kwargs: typing.Any,
) -> ui.Tag:
    id_num = len(slider_nums)
    slider_nums.append(id_num)
    return ui.row(
        ui.column(
            6,
            ui.input_slider(
                f"s{id_num}",
                label,
                **kwargs,
            ),
        ),
        ui.column(
            6,
            ui.output_text_verbatim(f"txt{id_num}", placeholder=True),
        ),
    )


app_ui = ui.page_fluid(
    ui.h2("Sliders!"),
    slider_row(
        "Integer",
        min=0,
        max=1000,
        value=500,
    ),
    slider_row(
        "Range",
        min=1,
        max=1000,
        value=(200, 500),
    ),
    slider_row(
        "Custom Format",
        min=0,
        max=10000,
        value=0,
        step=2500,
        pre="$",
        sep=",",
        animate=True,
        ticks=True,
    ),
    slider_row(
        "Looping Animation",
        min=1,
        max=2000,
        value=1000,
        step=10,
        animate=ui.AnimationOptions(interval=300, loop=True),
    ),
    slider_row(
        "Single Animation",
        min=0,
        max=5,
        value=1,
        step=1,
        animate=ui.AnimationOptions(interval=100, loop=False),
    ),
    # TODO-future; Test dates
)


def server(input: Inputs, output: Outputs, session: Session):
    def make_output(id_num: int):
        name = f"txt{id_num}"

        @output(id=name)
        @render.text
        def _():
            return input[f"s{id_num}"]()

    for id_num in slider_nums:
        make_output(id_num)


app = App(app_ui, server)



shiny_html_deps = "1.8.0.9000"
bslib = "0.6.1.9001"
htmltools = "0.5.7"
bootstrap = "5.3.1"
requirejs = "2.3.6"

__all__ = (
    "shiny_html_deps",
    "bslib",
    "htmltools",
    "bootstrap",
    "requirejs",
)



from conftest import ShinyAppProc, create_doc_example_core_fixture
from controls import InputSelectize
from playwright.sync_api import Page, expect

app = create_doc_example_core_fixture("input_selectize")


def test_input_selectize_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    state = InputSelectize(page, "state")

    expect(state.loc_label).to_have_text("Choose a state:")
    state.expect_label("Choose a state:")

    # TODO: This test was being run against input_select, not input_selectize
    # and none of these expectations pass. We need to add additional test methods
    # for the InputSelectize class to test this behaviour.
    # https://github.com/posit-dev/py-shiny/issues/1100

    # state.expect_choices(["NY", "NJ", "CT", "WA", "OR", "CA", "MN", "WI", "IA"])
    # state.expect_choice_labels(
    #     [
    #         "New York",
    #         "New Jersey",
    #         "Connecticut",
    #         "Washington",
    #         "Oregon",
    #         "California",
    #         "Minnesota",
    #         "Wisconsin",
    #         "Iowa",
    #     ]
    # state.expect_choice_groups(["East Coast", "West Coast", "Midwest"])

    # state.expect_multiple(True)

    # state.expect_width(None)

    # state.set("Iowa")

    # state.expect_selected("IA")



import pytest
from conftest import ShinyAppProc
from controls import (
    LayoutNavSetBar,
    LayoutNavSetCardPill,
    LayoutNavSetCardTab,
    LayoutNavSetCardUnderline,
    LayoutNavSetPill,
    LayoutNavSetPillList,
    LayoutNavsetTab,
    LayoutNavSetUnderline,
)
from playwright.sync_api import Page


@pytest.mark.skip_browser("webkit")
def test_nav(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # Update the page size to be wider
    page.set_viewport_size({"width": 1500, "height": 800})

    # make it in a funcn and then loop through it
    # navset_tab
    navset_tab = LayoutNavsetTab(page, "navset_tab")
    navset_tab.expect_nav_values(["a", "b", "c"])
    navset_tab.expect_value("a")
    navset_tab.expect_content("navset_tab(): tab a content")
    navset_tab.set("b")
    navset_tab.expect_value("b")
    navset_tab.expect_content("navset_tab(): tab b content")

    # # navset_pill
    navset_pill = LayoutNavSetPill(page, "navset_pill")
    navset_pill.expect_nav_values(["a", "b", "c"])
    navset_pill.expect_value("a")
    navset_pill.expect_content("navset_pill(): tab a content")
    navset_pill.set("b")
    navset_pill.expect_value("b")
    navset_pill.expect_content("navset_pill(): tab b content")

    # navset_underline
    navset_underline = LayoutNavSetUnderline(page, "navset_underline")
    navset_underline.expect_nav_values(["a", "b", "c"])
    navset_underline.expect_value("a")
    navset_underline.expect_content("navset_underline(): tab a content")
    navset_underline.set("b")
    navset_underline.expect_value("b")
    navset_underline.expect_content("navset_underline(): tab b content")

    # navset_card_tab
    navset_card_tab = LayoutNavSetCardTab(page, "navset_card_tab")
    navset_card_tab.expect_nav_values(["a", "b", "c"])
    navset_card_tab.expect_value("a")
    navset_card_tab.expect_content("navset_card_tab(): tab a content")
    navset_card_tab.set("b")
    navset_card_tab.expect_value("b")
    navset_card_tab.expect_content("navset_card_tab(): tab b content")

    # navset_card_pill
    navset_card_pill = LayoutNavSetCardPill(page, "navset_card_pill")
    navset_card_pill.expect_nav_values(["a", "b", "c"])
    navset_card_pill.expect_value("a")
    navset_card_pill.expect_content("navset_card_pill(): tab a content")
    navset_card_pill.set("b")
    navset_card_pill.expect_value("b")
    navset_card_pill.expect_content("navset_card_pill(): tab b content")

    # navset_card_underline
    navset_card_underline = LayoutNavSetCardUnderline(page, "navset_card_underline")
    navset_card_underline.expect_nav_values(["a", "b", "c"])
    navset_card_underline.expect_value("a")
    navset_card_underline.expect_content("navset_card_underline(): tab a content")
    navset_card_underline.set("b")
    navset_card_underline.expect_value("b")
    navset_card_underline.expect_content("navset_card_underline(): tab b content")

    # navset_pill_list
    navset_card_pill = LayoutNavSetPillList(page, "navset_pill_list")
    navset_card_pill.expect_nav_values(["a", "b", "c"])
    navset_card_pill.expect_value("a")
    navset_card_pill.expect_content("navset_pill_list(): tab a content")
    navset_card_pill.set("b")
    navset_card_pill.expect_value("b")
    navset_card_pill.expect_content("navset_pill_list(): tab b content")

    # Page_navbar
    navset_bar = LayoutNavSetBar(page, "page_navbar")
    navset_bar.expect_nav_values(["a", "b", "c"])
    navset_bar.expect_value("a")
    navset_bar.expect_content("page_navbar: tab a content")
    navset_bar.set("b")
    navset_bar.expect_value("b")
    navset_bar.expect_content("page_navbar: tab b content")
